#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009-2017 ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

# This code is used check job generation
#
import os
import shutil   # used to remove directory tree

import ecflow_test_util as Test
from ecflow import Defs, JobCreationCtrl, TaskVec, File, Client, debug_build

def ecf_includes() : return File.source_dir() + "/Pyext" + "/test/data/includes"

def create_defs(ecf_home,task_vec):
    defs = Defs();
    suite = defs.add_suite("suite_job_gen");
    suite.add_variable("ECF_HOME",ecf_home);
    suite.add_variable("ECF_CLIENT_EXE_PATH",File.find_client());
    suite.add_variable("SLEEPTIME","1");
    suite.add_variable("ECF_INCLUDE",ecf_includes());
    
    fam =  suite.add_family("family")
    t1 = fam.add_task("t1")
    t1.add_event("eventname");
    t1.add_meter("meter",0,100,100)
    
    t2 = fam.add_task("t2")
    t3 = fam.add_task("t3")
    t4 = fam.add_task("dummy_task")
    t4.add_variable("ECF_DUMMY_TASK","any"); # stop job generation errors for tasks without an .ecf
    task_vec.append(t1);
    task_vec.append(t2);
    task_vec.append(t3);
    task_vec.append(t4);
    
    return defs


def delete_jobs(task_vec, ecf_home):
    print("delete jobs")
    for task in task_vec:
        the_job_file = ecf_home + task.get_abs_node_path() + ".job" + task.get_try_no()
        if os.path.exists(the_job_file) :
            print("removing file " + the_job_file)
            try: os.remove(the_job_file)
            except: pass
        man_file = ecf_home + task.get_abs_node_path() + ".man"  
        if os.path.exists(man_file) :
            print("removing man_file " + man_file)
            try: os.remove(man_file)
            except: pass
   
def check_jobs(task_vec, ecf_home):
    print("Check job file exists")
    for task in task_vec:
        variable = task.find_variable("ECF_DUMMY_TASK")
        if not variable.empty(): continue;
        the_job_file = ecf_home + task.get_abs_node_path() + ".job" + task.get_try_no()
      
        if os.path.exists(the_job_file) :
            print("Found job file " + the_job_file)
        else:
            assert False, "Could not find job file " + the_job_file




if __name__ == "__main__":
    print("####################################################################")
    print("Running ecflow version " + Client().version() + " debug build(" + str(debug_build()) +")")
    print("####################################################################")
 
    workspace = File.source_dir();
    print(workspace)
    
    ecf_home = workspace + "/Pyext/test/data/ECF_HOME"
    task_vec = TaskVec()
    defs = create_defs(ecf_home,task_vec)
    print(str(defs))

    print("Generate jobs for *ALL* tasks, to default locations ECF_HOME/ECF_NAME.job0") 
    print(defs.check_job_creation())   
    check_jobs(task_vec,ecf_home)
    delete_jobs(task_vec,ecf_home)
       
    print("\nGenerate jobs for *ALL* tasks, to default locations ECF_HOME/ECF_NAME.job0")
    job_ctrl = JobCreationCtrl()
    defs.check_job_creation( job_ctrl )       
    print(job_ctrl.get_error_msg())
    check_jobs(task_vec,ecf_home)
    delete_jobs(task_vec,ecf_home)
    
    print("\nGenerate jobs for all nodes, under path, to default locations ECF_HOME/ECF_NAME.job0")    
    job_ctrl = JobCreationCtrl()
    job_ctrl.set_node_path( task_vec[0].get_abs_node_path() )    
    defs.check_job_creation(job_ctrl)       
    print(job_ctrl.get_error_msg());
    delete_jobs(task_vec,ecf_home)
   
    print("\nGenerate jobs for all tasks, to the specified directory")
    # Directory will automatically created under the provided directory
    job_ctrl = JobCreationCtrl()
    job_ctrl.set_dir_for_job_creation(workspace + "/Pyext/test/data")  # generate jobs file under this directory
    defs.check_job_creation(job_ctrl)
    print(job_ctrl.get_error_msg())
    
    generated_dir = job_ctrl.get_dir_for_job_creation() + "/suite_job_gen"
    print("removing directory tree " + generated_dir)
    shutil.rmtree(generated_dir)      
 
 
    # Create jobs for all task, to a TMP directory ($TMPDIR/ecf_check_job_creation/ECF_NAME.job0
    # When run via rsh then TMPDIR may not be defined, hence expect exception to be thrown
    try:
        job_ctrl = JobCreationCtrl()                    
        job_ctrl.generate_temp_dir()       
        defs.check_job_creation(job_ctrl)
        print(job_ctrl.get_error_msg())
        print("removing directory tree " + job_ctrl.get_dir_for_job_creation())
        shutil.rmtree(job_ctrl.get_dir_for_job_creation())     
    
        delete_jobs(task_vec,ecf_home)
        print("All test pass")
    except RuntimeError as e:
        print("failed: " + str(e))    



#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009-2017 ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

#  code for testing pointers and hierarchy in python

from ecflow import Suite, Family, Task, Defs, Client, debug_build

def create_defs(name=""):
    defs = Defs()
    defs.add_variable("DEFS_VAR","0")
    suite_name = name
    if len(suite_name) == 0: suite_name = "s1"
    suite = defs.add_suite(suite_name);
    suite.add_variable("SUITE_VAR","1")

    f1 = suite.add_family("f1")
    f1.add_task("f1_t1").add_variable("TASK_VAR","3")
    f1.add_task("f1_t2").add_variable("TASK_VAR","3")
    f2 = suite.add_family("f2")
    f2.add_task("f2_t1").add_variable("TASK_VAR","3")
    f2.add_task("f2_t2").add_variable("TASK_VAR","3")
    return defs;

if __name__ == "__main__":    
    print("####################################################################")
    print("Running ecflow version " + Client().version() + " debug build(" + str(debug_build()) +")")
    print("####################################################################")

    defs = create_defs()
    s1 = defs.find_suite("s1")
    assert s1 != None, "expected find find suite s1"
    assert "s1" in defs, "expected find find suite s1"

    sx = defs.find_suite("sx")
    assert sx == None, "expected not to find suite sx"
    
    f1 = s1.find_family("f1")
    assert f1 != None, "expected find find family f1"

    fx = s1.find_family("fx")
    assert fx == None, "expected not to find family fx"

    f2 = s1.find_family("f2")
    assert f2 != None, "expected find find family f2"
    
    f2_t1 = f2.find_task("f2_t1")
    assert f2_t1 != None, "Expected to find task"

    f2_t2 = f2.find_task("f2_t2")
    assert f2_t2 != None, "Expected to find task"

    f2_tx = f2.find_task("f2_tx")
    assert f2_tx == None, "Expected not to find task"

    tasks = defs.get_all_tasks()
    assert len(tasks) == 4, "Expected four tasks, but found " + str(len(tasks))
    for task in tasks:
        # test find variable
        var = task.find_parent_variable("TASK_VAR")
        assert not var.empty()
        assert var.value() == "3"

        var = task.find_variable("TASK_VAR")
        assert not var.empty()
        assert var.value() == "3"
        
        var = task.find_parent_variable("SUITE_VAR");
        assert not var.empty()
        assert var.value() == "1"
         
        var = task.find_parent_variable("DEFS_VAR")
        assert not var.empty()
        assert var.value() == "0"

        var = task.find_parent_variable("MADE_UP_VAR")
        assert var.empty()

        if task.name() == "f1_t1":
            node = task.find_node_up_the_tree("f1")
            assert node != None
            assert node.get_abs_node_path() == "/s1/f1"
            
            node = task.find_node_up_the_tree("s1")
            assert node != None
            assert node.get_abs_node_path() == "/s1"

            node = task.find_node_up_the_tree("freddd")
            assert node == None
            
    print("All Tests pass")



import os
import ecflow

print("PYTHONPATH=====================================================")
print(os.environ['PYTHONPATH'].split(os.pathsep))

print("Creating Client")
ci = ecflow.Client()
ci.set_host_port("%ECF_HOST%","%ECF_PORT%")
ci.set_child_pid(os.getpid())
ci.set_child_path("%ECF_NAME%")
ci.set_child_password("%ECF_PASS%")
ci.set_child_try_no(%ECF_TRYNO%)

print("Only wait 5  minutes, if the server cannot be contacted (note default is 24 hours) before failing")
ci.set_child_timeout(300) 

try:
   ci.child_init();
   print("child init ok")
except RuntimeError as e:
   print("Error: abort in head.h " + str(e))
   ci.child_abort("Abort in head" + str(e))
except:
   print("Error: abort in head.h")
   ci.child_abort("Abort in head")



#!/usr/bin/env python
""" tkinter example with ecflow
"""

import Tkinter as tki
import Queue
import sys
import time
import os
import pwd
from threading import Thread
try: from scrolledlist import ScrolledList
except: 
    loc = "http://infohost.nmt.edu/tcc/help/lang/python/examples/scrolledlist"
    raise Exception("wget %s/scrolledlist.py; # Thanks NMT")

loc = "/usr/local/apps/ecflow/current/lib/python2.7/site-packages/ecflow"
sys.path.append(loc)
loc = "/usr/local/lib/python2.7/site-packages/ecflow"
sys.path.append(loc)
import ecflow as ec

PROGRAM_NAME = "ecflowview-overview"
BUTTON_FONT = ('times', 12)
MONO_FONT = ('lucidatypewriter', 14, 'bold')
DEBUG = 0
COLORS = {"aborted": "red",
          "active":  "green",
          "submitted": "cyan",
          "complete": "yellow",
          "suspended": "orange",
          "queued": "blue",
          "unknwon": "grey"}
running = [True]


class Label(object):
    """ a class to encapsulate what was a global variable"""
    inst = None

    def __init__(self, item): Label.inst = item

    @classmethod
    def update(cls):
        if Label.inst is None:
            return
        Label.inst.set(time.strftime("%a, %d %b %Y %H:%M:%S"))


class MenuBar(tki.Frame):
    """ guess"""

    def __init__(self, parent):
        tki.Frame.__init__(self, parent)
        self.__helpButton = self.__createHelp()
        self.__helpButton.grid(row=0, column=3)
        self.__updateButton = tki.Button(self, text='Update', font=BUTTON_FONT,
                                         command=parent.update)
        self.__updateButton.grid(row=0, column=2)

    def __createHelp(self):
        mb = tki.Menubutton(self, font=BUTTON_FONT,
                            relief=tki.RAISED, text='Help')
        menu = tki.Menu(mb)
        mb['menu'] = menu
        url = "https://software.ecmwf.int/wiki/display/ECFLOW/Documentation"

        def url1(): self.__url(url=url)

        def url2(): self.__url(url="http://effbot.org/tkinterbook/")
        menu.add_command(command=url1, label="confluence tutorial?")
        menu.add_command(command=url2, label="tkinter?")
        return mb

    def __url(self, url=None):
        if url is None:
            return
        os.system("firefox " + url)


class TaskList(tki.Frame):
    """ a class to host all tasks"""
    WIDTH = 40
    LINES = 80

    def __init__(self, parent, kind):
        tki.Frame.__init__(self, parent)
        self.__kind = kind
        self.__callback = None
        self.__label = tki.Label(
            self, font=BUTTON_FONT, background=COLORS[kind], text=kind)
        self.__label.grid(row=0, column=0, sticky=tki.W)
        self.__scrolledList = ScrolledList(
            self, width=self.WIDTH, height=self.LINES,
            callback=self.__callback)
        self.__scrolledList.grid(row=1, column=0)

    def insert(self, path):  self.__scrolledList.append(path)

    def clear(self):         self.__scrolledList.clear()


class PaceKeeper():
    """ get regular updates"""
    PACE = 60

    def __init__(self, item, queue):
        self._item = item
        Thread(target=self.process, args=(queue, running)).start()

    def process(self, queue, running):
        while running:
            queue.put(self._item.update)
            time.sleep(self.PACE)

    def run(self): self.update()

    def update(self, verbose=False):
        while True:
            self._item.update()
            time.sleep(self.PACE)


class Client(object):
    """ a class to focus on client-ecFlow-server comm"""

    def __init__(self, one="local-localhost@31415"):
        try:
            nick, hhh = one.split("-")
        except:
            hhh = one
            nick = None
        try:
            host, port = hhh.split("@")
        except:
            host = "localhost"
            port = 31415

        if nick is None:
            self.nick = "%s@%d" % (host, port)
        print "# client creation", nick, host, port
        self.nick = nick
        self.client = ec.Client(host, port)

    def process(self, win):
        Label.update()
        self.client.sync_local()
        defs = self.client.get_defs()
        if defs is None:
            print("# %s-%: empty content" % (self.host, self.port))
        Label.update()
        for suite in defs.suites:
            self.process_nc(suite, win)

    def process_nc(self, node, win):
        for item in node.nodes:
            if isinstance(item, ec.Task):
                self.process_node(item, win)
            else:
                self.process_nc(item, win)

    def process_node(self, node, wins):
        for kind, win in wins.items():
            status = "%s" % node.get_state()
            if status != kind:
                continue
            win.insert("%s:%s" % (self.nick, node.get_abs_node_path()))

try:
    sys.path.append('/usr/local/apps/sms/lib/')
    import cdp
    NOCDP = 0
except:
    NOCDP = 1


class ClientSMS(Client):
    """ example of class derivation to connect another server-scheduler"""
    init = 0

    def __init__(self, host="localhost", port=31415, nick=None):
        if NOCDP:
            return
        if not ClientSMS.init:
            cdp.cdp_init([""], 0)
            ClientSMS.init = 1
        else:
            pass
        node = cdp.sms_node()
        tree = cdp.sms_info_pointed(node)
        cdp.sms_numbers(port, 0)
        timeout = 60
        passwd = "1"
        user = get_username()

        if DEBUG:
            print "#MSG: login:", nick, host, port, user
        handle, rc = cdp.sms_client_login(host, user, passwd,
                                          timeout, port, tree)
        self.handle = handle
        self.host = host
        self.port = port
        self.num = -1
        self.nick = nick
        if nick is None:
            self.nick = "%s@%d" % (self.host, self.port)

    def process_node(self, node, wins):
        if node.type != cdp.NODE_TASK:
            kid = node.kids
            while kid:
                self.process_node(kid, wins)
                kid = kid.next
        else:
            for kind, win in wins.items():
                state = "%s" % ClientSMS.status(node)
                if state != kind:
                    continue
                path = cdp.sms_node_full_name(node)
                win.insert("%s:%s" % (self.nick, path))
                if DEBUG:
                    print self.host, self.port, path, state

    def process(self, win):
        rc = cdp.sms_client_news(self.handle, self.num)
        if rc or self.num == -1:
            Label.update()
            state = cdp.sms_client_status2(self.handle, self.num)
            cdp.sms_client_status(self.handle, state, self.num)

            item = cdp.sms_status2node(state)
            if not item:
                return
            self.process_node(item, win)
            Label.update()
            cdp.sms_node_free(item)

    @classmethod
    def status(cls, node):
        s_status = {
            -1: "unknown",
            0: "unknown",
            1: "suspended",
            2: "complete",
            3: "queued",
            4: "submitted",
            5: "active",
            6: "aborted",
            7: "shutdown",
            8: "halted",
            9: "unknown", }
        return s_status[node.status]

    def __exit__(self):
        cdp.sms_client_logout(self.handle)
        print "# logout from", self.nick

    def __delete__(self):
        cdp.sms_client_logout(self.handle)
        print "# logout from", self.nick


class Application(tki.Frame):
    """ main """

    def __init__(self, master=None, client=None, queue=None):
        tki.Frame.__init__(self, master)
        if client is None:
            self.__clients = [Client("localhost@31415"), ]
        elif type(client) == set:
            self.__clients = client
        else:
            self.__clients = [client]

        self.__queue = queue
        width = 640
        height = 780
        self.canvas = tki.Canvas(width=width, height=height, bg='black')
        self.grid()
        self.createWidgets()
        self.canvas.after(50, self.check_queue)

    def createWidgets(self):
        rowx = 1
        glob = Label(tki.StringVar(self))
        root = self
        self.__menuBar = MenuBar(root)
        self.__menuBar.grid(row=0, column=0, sticky=tki.W)
        self.label = tki.Label(root, textvariable=Label.inst)
        self.label.grid(row=0, column=2, sticky=tki.E)
        self.__wins = dict()
        rowx += 1
        colx = 0
        kinds = ("active", "aborted", "submitted")
        for kind in kinds:
            self.__wins[kind] = TaskList(root, kind)
            self.__wins[kind].grid(row=rowx, column=colx,
                                   sticky=tki.S + tki.E + tki.W)
            colx += 1
        self.update()

    def check_queue(self):
        try:
            self.__queue.get(block=False)
        except Queue.Empty:
            pass
        else:
            self.update()
        self.canvas.after(50, self.check_queue)

    def update(self):
        Label.update()
        for kind, win in self.__wins.items():
            win.clear()
        for client in self.__clients:
            if type(client) == list:
                for clt in client:
                    clt.process(self.__wins)
            else:
                try:
                    client.process(self.__wins)
                except:
                    pass


def get_username(): return pwd.getpwuid(os.getuid())[0]


def get_uid(): return pwd.getpwnam(get_username()).pw_uid


if __name__ == '__main__':
    try:
        port = 1500 + int(get_uid())
    except:
        port = 31415

    if len(sys.argv) > 0:
        clients = []
        for num in xrange(1, len(sys.argv)):
            clients.append(Client(sys.argv[num]))
    else:
        clients = [Client("localhost%d" % port)]

    queue = Queue.Queue()
    # app = AppSms(host, port, queue)
    app = Application(client=clients, queue=queue)
    app.master.title(PROGRAM_NAME)
    app.columnconfigure(0, weight=1)
    app.rowconfigure(0, weight=1)

    PaceKeeper(app, queue)
    app.mainloop()

"""
python ./overview.py localhost-localhost@31415
"""



#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009-2017 ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
#  code for testing client code in python
import time
import os
import pwd
import sys  # determine python version
from datetime import datetime
import shutil   # used to remove directory tree

from ecflow import Defs, Clock, DState,  Style, State, PrintStyle, File, Client, SState, File, debug_build
import ecflow_test_util as Test


def ecf_includes() : return File.source_dir() + "/Pyext" + "/test/data/python_includes"

def create_defs(name,the_port):
    defs = Defs()
    suite_name = name
    if len(suite_name) == 0: suite_name = "s1"
    suite = defs.add_suite(suite_name);
    
    ecfhome = Test.ecf_home(the_port);
    suite.add_variable("ECF_HOME", ecfhome);
    suite.add_variable("ECF_INCLUDE", ecf_includes());

    family = suite.add_family("f1")
    # sys.version_info is a tuple containing (major,minor,micro,releaselevel,serial)
    # releaselevel = alpha beta candidate final
    if (sys.version_info > (3, 0)):
        family.add_variable("ECF_JOB_CMD","python3 %ECF_JOB% 1> %ECF_JOBOUT% 2>&1")
    else:
        family.add_variable("ECF_JOB_CMD","python %ECF_JOB% 1> %ECF_JOBOUT% 2>&1")

    task = family.add_task("t1")
    task.add_event("event_fred")
    task.add_meter("meter", 0, 100)
    task.add_label("label_name", "value")

    family.add_task("t2")  # test wait
 
    return defs;
    

def test_client_run(ci):            
    print("\ntest_client_run " + ci.get_host() + ":" + str(ci.get_port()))
    print(" ECF_HOME(" + Test.ecf_home(ci.get_port()) + ")")
    print(" ECF_INCLUDES(" + ecf_includes() + ")")
    ci.delete_all()     
    defs = create_defs("test_client_run",ci.get_port())  
    suite = defs.find_suite("test_client_run")
    suite.add_defstatus(DState.suspended)

    # create the ecf file /test_client_run/f1/t1
    ecf_home = Test.ecf_home(ci.get_port())
    dir = ecf_home + "/test_client_run/f1"
    
    # on cray creating recursive directories can fail, try again. yuk
    try:
        if not os.path.exists(dir): os.makedirs(dir)
    except:
        try:
            if not os.path.exists(dir): os.makedirs(dir)
        except:
            # try breaking down
            if not os.path.exists(ecf_home): os.makedirs(ecf_home)
            new_dir = ecf_home + "/test_client_run"
            if not os.path.exists(new_dir): os.makedirs(new_dir)
            new_dir = ecf_home + "/test_client_run/f1"
            if not os.path.exists(new_dir): os.makedirs(new_dir)
    if not os.path.exists(dir): os.makedirs(dir)

    file = dir + "/t1.ecf"
    contents = "%include <head.py>\n\n"
    contents += "print('doing some work')\n"
    contents += "try:\n"
    contents += "    ci.child_event('event_fred')\n"
    contents += "    ci.child_meter('meter',100)\n"
    contents += "    ci.child_label('label_name','100')\n"
    contents += "    print('Finished event,meter and label child commands')\n"
    contents += "except:\n"
    contents += "    ci.child_abort()\n\n"
    contents += "%include <tail.py>\n"
    open(file,'w').write(contents)
    print(" Created file " + file)
      
    # create the ecf file /test_client_run/f1/t2
    file = dir + "/t2.ecf"
    contents = "%include <head.py>\n\n"
    contents += "print('Waiting for /test_client_run/f1/t1 == complete')\n"
    contents += "try:\n"
    contents += "    ci.child_wait('/test_client_run/f1/t1 == complete')\n"
    contents += "    print('Finished waiting')\n"
    contents += "except:\n"
    contents += "    ci.child_abort()\n\n"
    contents += "%include <tail.py>\n"
    open(file,'w').write(contents)
    print((" Created file " , file))
      
    ci.restart_server()
    ci.load(defs)           
    ci.begin_all_suites()
    ci.run("/test_client_run", False)
    print(" Running the test, wait for suite to complete ...")  

    count = 0
    while 1:
        count += 1
        ci.sync_local() # get the changes, synced with local defs
        suite = ci.get_defs().find_suite("test_client_run")
        assert suite != None, " Expected to find suite test_client_run:\n" + str(ci.get_defs())
        if suite.get_state() == State.complete:
            break;
        if suite.get_state() == State.aborted:
            print(defs);
            assert False," Suite aborted \n"  
        time.sleep(2)
        if count > 20:
            assert False, " test_client_run aborted after " + str(count) + " loops:\n" + str(ci.get_defs())
        
    ci.log_msg("Looped " + str(count) + " times")
    
    if not Test.debugging():
        dir_to_remove = Test.ecf_home(ci.get_port()) + "/" + "test_client_run"
        print((" Test OK: removing directory " , dir_to_remove))
        shutil.rmtree(dir_to_remove)      
        
if __name__ == "__main__":
    print("####################################################################")
    print("Running ecflow version " + Client().version() + " debug build(" + str(debug_build()) +")")
    print("####################################################################")

    with Test.Server() as ci:
        PrintStyle.set_style( Style.STATE ) # show node state 
        test_client_run(ci)  
        print("\nAll Tests pass ======================================================================")    



#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009-2017 ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

#  code for testing derivation works
import ecflow

class MyDefs(ecflow.Defs): pass
class MySuite(ecflow.Suite): pass
class MyFamily(ecflow.Family): pass
class MyTask(ecflow.Task): pass
class MyClient(ecflow.Client): pass

if __name__ == "__main__":
    print("####################################################################")
    print("Running ecflow version " + ecflow.Client().version() + " debug build(" + str(ecflow.debug_build()) +")")
    print("####################################################################")
 
    print("All tests pass")


#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009-2017 ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
# Utility code used in the tests.
#
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
from socket import gethostname 
import os,sys,fnmatch
import fcntl
import datetime,time
import shutil   # used to remove directory tree

from ecflow import Client, debug_build, File

def all_files(root, patterns='*', single_level=False, yield_folders=False):
    """Expand patterns from semi-colon separated string to list"""
    patterns = patterns.split(';')
    for path, subdirs, files in os.walk(root):
        if yield_folders:
            files.extend(subdirs)
        files.sort()
        for name in files:
            for pattern in patterns:
                if fnmatch.fnmatch(name,pattern):
                    yield os.path.join(path, name)
                    break
        if single_level:
            break    
        
# Enable to stop data being deleted, and stop server from being terminated
def debugging() : return False

def ecf_home(port): 
    # debug_build() is defined for ecflow. Used in test to distinguish debug/release ecflow
    # Vary ECF_HOME based on debug/release/port allowing multiple invocations of these tests
    if debug_build():
        return os.getcwd() + "/test/data/ecf_home_debug_" + str(port)
    return os.getcwd() + "/test/data/ecf_home_release_" + str(port)

def get_parent_dir(file_path):
    return os.path.dirname(file_path)

def log_file_path(port): return "./" + gethostname() + "." + port + ".ecf.log"
def checkpt_file_path(port): return "./" + gethostname() + "." + port + ".ecf.check"
def backup_checkpt_file_path(port): return "./" + gethostname() + "." + port + ".ecf.check.b"
def white_list_file_path(port): return "./" + gethostname() + "." + port + ".ecf.lists"

def clean_up_server(port):
    print("   clean_up " + port)
    try: os.remove(log_file_path(port))
    except: pass
    try: os.remove(checkpt_file_path(port))
    except: pass
    try: os.remove(backup_checkpt_file_path(port))
    except: pass
    try: os.remove(white_list_file_path(port))  
    except: pass
    
def clean_up_data(port):
    print("   Attempting to Removing ECF_HOME " + ecf_home(port))
    try: 
        shutil.rmtree(ecf_home(port),True)   # True means ignore errors 
        print("   Remove OK") 
    except: 
        print("   Remove Failed") 
        pass
        
# =======================================================================================
class EcfPortLock(object):
    """allow debug and release version of python tests to run at the same
    time, buy generating a unique port each time"""
    def __init__(self):
        print("   EcfPortLock:__init__")
        pass
    
    def find_free_port(self,seed_port):
        print("   EcfPortLock:find_free_port starting with " + str(seed_port))
        at_time = datetime.datetime.fromtimestamp(time.time()).strftime('%H:%M:%S')
        port = seed_port
        while 1:
            if self._free_port(port) == True:
                print("   *FOUND* free server port " + str(port) + " : " + at_time)
                if self._do_lock(port) == True:
                    break;
            else:
                print("   *Server* port " + str(port) + " busy, trying next port " + at_time)
            port = port + 1
            
        return str(port)  
    
    def _free_port(self,port):
        try:
            ci = Client()
            ci.set_host_port("localhost",str(port))
            ci.ping() 
            return False
        except RuntimeError as e:
            return True
            
    def _do_lock(self,port):
        file = self._lock_file(port)
        at_time = datetime.datetime.fromtimestamp(time.time()).strftime('%H:%M:%S')
        if os.path.exists(file):
            print("   *LOCKED* lock file exists " + file + " : " + at_time )
            return False
        try:
            fp = open(file, 'w') 
            try:
                self.lock_time = at_time
                fcntl.lockf(fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
                self.lock_file_fp = fp
                print("   *LOCKED* file " + file + " : " + self.lock_time )
                return True;
            except IOError:
                print("   Could *NOT* lock file " + file + " trying next port : " + at_time)
                return False
        except IOError as e:
             print("   Could not open file " + file + " for write trying next port : " + at_time)
             return False
        
    def remove(self,port):
        release_time = datetime.datetime.fromtimestamp(time.time()).strftime('%H:%M:%S')
        file = self._lock_file(port)
        print("   Remove lock file : " + file + " : lock_time: " + self.lock_time + " release_time: " + release_time)
        self.lock_file_fp.close()
        os.remove(file)
    
    def _lock_file(self,port):
        if "ECF_PORT_LOCK_DIR" in os.environ:
            lock_file = os.environ["ECF_PORT_LOCK_DIR"] + "/" + str(port) + ".lock"
            #print("     EcfPortLock::_lock_file ECF_PORT_LOCK_DIR: " + lock_file)
            return lock_file
        if os.path.exists(File.source_dir()):
            lock_file = File.source_dir() + "/" + str(port) + ".lock"
            #print("     EcfPortLock::_lock_file File::source_dir(): " + lock_file)
            return lock_file
        lock_file = os.getcwd() + "/" + str(port) + ".lock"
        #print("     EcfPortLock::_lock_file os.getcwd(): " + lock_file)
        return lock_file
        
# ===============================================================================

class Server(object):
    """TestServer: allow debug and release version of python tests to run at the same
    time, by generating a unique port each time"""
    def __init__(self):
        print("Server:__init__: Starting server")      
        if not debugging():
            seed_port = 3153
            if debug_build(): seed_port = 3152
            self.lock_file = EcfPortLock()
            self.the_port = self.lock_file.find_free_port(seed_port)   
        else:
            self.the_port = "3152"
     
        # Only worth doing this test, if the server is running
        # ON HPUX, having only one connection attempt, sometimes fails
        #ci.set_connection_attempts(1)     # improve responsiveness only make 1 attempt to connect to server
        #ci.set_retry_connection_period(0) # Only applicable when make more than one attempt. Added to check api.
        self.ci = Client("localhost", self.the_port)
     
    def __enter__(self):
        try:
            st = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
            print("Server:__enter__: About to ping localhost: " + self.the_port +  " : " + st)       
            self.ci.ping() 
            print("   ------- Server all ready running on port " + self.the_port + " *UNEXPECTED* ------")
            sys.exit(1)
        except RuntimeError as e:
            print("   ------- Server *NOT* running on port " + self.the_port + " as *EXPECTED* ------ ") 
            print("   ------- Start the server on port " + self.the_port + " ---------")  
            clean_up_server(str(self.the_port))
            clean_up_data(str(self.the_port))
    
            server_exe = File.find_server();
            assert len(server_exe) != 0, "Could not locate the server executable"
        
            server_exe += " --port=" + self.the_port + " --ecfinterval=4 &"
            print("   TestClient.py: Starting server ", server_exe)
            os.system(server_exe) 
        
            print("   Allow time for server to start")
            if self.ci.wait_for_server_reply() :
                print("   Server has started")
            else:
                print("   Server failed to start after 60 second !!!!!!")
                assert False , "Server failed to start after 60 second !!!!!!"
            
        print("   Run the tests, leaving Server:__enter__:") 

        # return the Client, that can call to the server
        return self.ci
    
    def __exit__(self,exctype,value,tb):
        print("Server:__exit__: Kill the server, clean up log file, check pt files and lock files, ECF_HOME")
        print("   exctype:",exctype)
        print("   value:",value)
        print("   tb:",tb)
        print("   Terminate server")
        self.ci.terminate_server()  
        print("   Terminate server OK")
        self.lock_file.remove(self.the_port)
        clean_up_server(str(self.the_port))
        
        # Do not clean up data, if an assert was raised. This allow debug
        if exctype == None:
            clean_up_data(str(self.the_port))
        return False
        
  


#!/usr/bin/env python
from __future__ import print_function
# -m trace --count -C .
"""an example for a simple python client to help ecFlow scheduling
 + unit test
 + task loading in the suite
$manual

  MIRRORING suite from server to server

  test - suspend suite and rerun job if aborted... Thanks

$end
$comment
  a comment in a comment
$end

"""
import getopt
import os
import pwd
import sys
import time
import unittest
from time import gmtime, strftime

ECFLOWP = "/usr/local/apps/ecflow/current"
ECFLOWC = ECFLOWP + "/bin/ecflow_client"
sys.path.append(ECFLOWP + "/lib/python2.7/site-packages/ecflow")
child = None


def getreq(): return Family("getreq").add(Task("collectreq"))


def ymd():
    return (Edit("YMD", "29090909"), Label("ymd", "29090909"))


def excepthook(exctype, value, traceback):
    if exctype == KeyboardInterrupt:
        if child:
            child.report("abort", "keyb")
    # elif exctype == NameError:
    #     if child: child.report("abort", "name")
    else:
        sys.__excepthook__(exctype, value, traceback)
        if child:
            child.report("abort", "gen")


def fullname(item):
    if type(item) == cdp.sms_node:
        if item.name != '/':
            return fullname(item.parent) + '/' + item.name
    return ""


sys.excepthook = excepthook
CDP = None

from ecf import *  # mirror
import ecf
import ecflow


class Edit(ecf.Variables):
    pass


MICRO = "$$"  # keep ecFlow pleased with micro character balance
DEBUG = 1
DEBUG = 0
USER = "emos"
XPING = "/usr/local/apps/sms/bin/smsping"
x_status = {-1: "unknown",
              0: "unknown",
              1: "suspended",
              2: "complete",
              3: "queued",
              4: "submitted",
              5: "active",
              6: "aborted",
              7: "shutdown",
              8: "halted",
              9: "unknown", }
x_type = {13: "definition",
            12: "suite",
            11: "family",
            10: "task",
            32: "alias", }


############################
def usage():
    print(sys.argv[0], """
      -h # help
      -m <server@port> # mirror path fri
      -t # unit test
      -r # load the task definition in a suite, associated with -p option
      -r # with -m, allow suite loading in destination

can also be used as a module: gen_task, gen_suite

    """)
    for num, val in enumerate(sys.argv):
        print(num, val)
    child.report("abort")
    sys.exit(2)


############################
def get_uid():
	return pwd.getpwnam(get_username()).pw_uid


############################
def gen_task(var=True, load_only=False, kind=None):
    from ecf import Task, Label, Variables
    pwd = os.getcwd()
    name = sys.argv[0]
    if '/' in name:
        name = name.split('/')[-1]
    if '.' in name:
        name = name.split('.')[0]
    rec = "00:30"
    if kind is None:
        pass
    cmd = "trimurti $USER$ $WSHOST$ $ECF_JOB$ $ECF_JOBOUT$"
    WDIR = pwd
    ODIR = pwd

    load = (Label("info", "mirror suites..."),
            Variables(ECF_MICRO=MICRO[0],
                      ECF_FILES=WDIR,
                      ECF_HOME=ODIR,
                      ECF_EXTN=".py",
                      ECF_JOB_CMD=cmd, ))
    if load_only:
        return load
    return Task(name).add(
        Cron("04:00 20:00 " + rec),
        If(var, load))


############################
# def stop(msg, num, child=None):
#     print(msg);
#     if child:
#         child.report(msg)
#         if num == 0: child.report("complete")
#     sys.stdout.flush()
#     sys.stderr.flush()
#     sys.exit(num)
def timer(a, b):
    child.report("timeout")
    child.report("complete")


import signal
signal.signal(signal.SIGALRM, timer)
signal.alarm(180)


############################
class Child(object):
    """ facilitate job behaviour with ecFlow server:
    just instanciate and call complete when finished

    obj = Child(); obs.report("complete")

    this does nothing when script is called from command line
    """

    def __init__(self):
        import signal
        env = {"ECF_HOST": "$ECF_HOST:$",
               "ECF_NODE": "$ECF_NODE:$",
               # check can be on a server, child on another
               "ECF_PASS": "$ECF_PASS$",
               "ECF_NAME": "$ECF_NAME$",
               "ECF_PORT": "$ECF_PORT:0$",
               "ECF_TRYNO": "$ECF_TRYNO$", }
        self.client = None
        if MICRO[0] in env["ECF_PORT"]:
            print("#MSG: cli mode")
            return
        print("#MSG: will communicate with server...")
        print("#MSG: kill: ssh %s kill -15 %d" % (os.uname()[1], os.getpid()))
        self.client = ecflow.Client()
        if MICRO[0] == MICRO:
            host = env["ECF_HOST"]
            print(host)
            if host == "" or "%" in host:
                host = env["ECF_NODE"]
            self.client.set_host_port(host, int(env["ECF_PORT"]))
            self.client.set_child_pid(os.getpid())
            self.client.set_child_path(env["ECF_NAME"])
            self.client.set_child_password(env["ECF_PASS"])
            self.client.set_child_try_no(int(env["ECF_TRYNO"]))
        else:
            host = "$ECF_HOST:$"
            print(host)
            if not (host != "" and "%" not in host):
                host = "$ECF_NODE:$"
            self.client.set_host_port(host, int("$ECF_PORT:0$"))
            self.client.set_child_pid(os.getpid())
            self.client.set_child_path("$ECF_NAME$")
            self.client.set_child_password("$ECF_PASS$")
            self.client.set_child_try_no(int("$ECF_TRYNO$"))
        self.client.child_init()

        self.client.child_label("info", "%s" % strftime("%Y-%m-%d %H:%M:%S",
                                                        gmtime()))
        self.client.set_child_timeout(20)
        print(self.client)

        for sig in (signal.SIGINT,
                    signal.SIGHUP,
                    signal.SIGQUIT,
                    signal.SIGILL,
                    signal.SIGTRAP,
                    signal.SIGIOT,
                    signal.SIGBUS,
                    signal.SIGFPE,
                    signal.SIGUSR1,
                    signal.SIGUSR2,
                    signal.SIGPIPE,
                    signal.SIGTERM,
                    signal.SIGXCPU,
                    signal.SIGPWR):
            signal.signal(sig, self.signal_handler)

    def signal_handler(self, signum, frame):
        """ catch signal """
        print('Aborting: Signal handler called with signal ', signum)
        if self.client:
            self.client.child_abort(
                "Signal handler called with signal " + str(signum))

    def __exit__(self, exc_type, exc_value, traceback):
        if self.client:
            self.client.child_abort()

    def report(self, msg, meter=None):
        """ communicate with ecFlow server """
        if not self.client:
            print("#MSG: no child")
            if msg in ("stop", "complete"):
                sys.stdout.flush()
                sys.stderr.flush()
                print("#MSG: no child stop")
                sys.exit(0)
            else:
                print(msg, meter)
            return
        elif msg in ("stop", "complete"):
            self.client.child_complete()
            self.client = None
            sys.stdout.flush()
            sys.stderr.flush()
            print("#MSG: stop")
            sys.exit(0)
        elif msg in ("abort", ):
            self.client.child_abort()
            self.client = None
            if 1:
                raise Exception(msg)
        elif meter:
            self.client.child_meter(msg, meter)
        else:
            self.client.child_label("info", msg)


############################
def gen_suite(host=None, port=None, path=None):
    import ecf
    from ecf import (Task, Label, Variables, Suite, Defs,
                     Family, Trigger, Defstatus, Meter, Event)
    destinations = dict()
    defs = Defs()
    if 1:
        dest = "localhost@3141"
        if path:
            if path[0] == '/':
                sname = path.split('/')[1]
                print(sname)
                if sname in definitions.keys():
                    defs.add_suite(
                        definitions[sname].add(
                            Defstatus("suspended")))
                    return defs

    sname = "mirror"
    if DEBUG:
        print(definitions.keys())
    defs.add_suite(
        Suite(sname).add(
            ecf.Defstatus("suspended"),
            Limit("one", 2),
            Inlimit("one"),
            gen_task(load_only=True),
            Variables(ALL_ECF="localhost:31415",
                      ALL_SMS="",
            [Family(name).add(
                Variables(DESTINATIONS=destinations[name]),
                Label("info", ""),
                gen_task(0, kind=name), )
             for name in definitions.keys()])))
    if DEBUG:
        print(defs)
    return defs


def o5hres(name="gsup"): return Family(name).add(
    Family("an").add(
        Family("main").add(
            Family("ed00").add(
                Label("info", "YMD"),
                Edit("YMD", "21000101"),  # USE_YMD
                Task("an"),
                Task("fc")))))


definitions = {  # strings as path for all nodes below sync
    # or suites, subset of the original, leaves are task...
    # or else not sync'ed
    # shall we add list of string/path ???
    "aaa": "/aaa/run/ext",

    "test": Suite("test").add(
        Defstatus("suspended"),
        gen_fcgroup1(name="abcd")), }
############################


class Mirror(object):

    """
  $nopp

  $end
    """

    def _protect_oper(self, server):
        if server == "xxx@1234":
            if 1:
                child.report("abort", "protect")
            elif ("/o" in self.path
                  or self.path == "all"
                  or "/eda" in self.path
                  or "/mc" in self.path):
                raise Exception("wont do that...")

    def _warmup(self):
        self._set_register(self.path)
        self.source.ch_register(False, self.register)
        self.source.sync_local()

    def __init__(self, server="eurus@1630", path="/e_41r2", replay=False):
        self.replay = replay  # allow replay from skel definition
        self.act = 1         # active ie force change allowed
        self.snap = 1        # sleep interval
        self.wait = 0  # refresh interval,continue as long as sub/act node exist
        self.path = path

        self._set_register(path)

        if server == "all":
            self.servers = "$DESTINATIONS$".split(" ")
        elif " " in server:
            self.servers = server.split(" ")
        elif "@" in server:
            host, port=server.split("@")
            self.servers=("%s@%s" % (host, port), )
        elif ":" in server:
            host, port=server.split(":")
            self.servers=("%s:%s" % (host, port), )

        host=os.getenv("ECF_HOST", "$ECF_HOST:none$")
        if MICRO[0] in host or host == "none":
            host = os.getenv("ECF_NODE", "$ECF_NODE$")
        # if MICRO[0] in host: host = "localhost"
        port = os.getenv("ECF_PORT", "$ECF_PORT$")
        if MICRO[0] in port:
            port = 1500 + int(get_uid())
        # when started with ecflow_start.sh
        # aka reference/source  server name
        self.sname = "%s:%s" % (host, port)

        if "$SOURCE:$" != "":
            self.sname = "$SOURCE:$"
        if DEBUG:
            print("#DBG: source is", self.sname)
        self.source = ecflow.Client(self.sname)
        self._warmup()

    def _set_register(self, path):
        if '/' == path[0]:
            suite = str(path.split('/')[1])
            self.register = [suite, ]

        elif path == "all":
            self.register = skel.keys()

        else:
            child.report("abort", "register")

    def _check(self, server, client, stop=0):
        if client == self.source:
            stop = 1

        for suite in self.register:
            top = client.get_defs().find_abs_node("/" + suite)

            if top is None:
                print("#WAR: node not found", suite, server, )
                if stop:
                    child.report("abort", "check")
                elif self.act and server != self.sname:
                    self._protect_oper(server)
                    if self.replay:
                        if type(definitions[suite]) == str:
                            print("#WAR: not yet")
                            child.report("abort", "str")
                        print("replaying")

                        defs = ecflow.Defs()
                        defs.add_suite(definitions[suite])
                        client.load(defs)
                        client.suspend('/' + suite)
                        client.begin_suite(suite)
                        client.sync_local()
                    else:
                        print("replay?")
                else:
                    pass

            elif server != self.sname and not (
                    top.is_suspended()  # normal backup server
                    or "%s" % top.get_state() == "unknown"):  # test server
                print("#ERR: top node shall be suspended - NOGO")
                print("#ERR:", server, top.name(), top.get_state())
                if 1:
                    child.report("abort", "state")
                print("#MSG: stop")
                child.report("abort", "state")  # sys.exit(1)

            else:
                pass

            if type(definitions[suite]) == str:
                path = definitions[suite]

            elif suite in self.path:
                path = self.path

            else:
                return

            node = client.get_defs().find_abs_node(path)
            if node:
                return
            print("#WAR: node not found", path, server)
            if stop:
                child.report("abort", "stop")

    def update_var(self, var, nfrom, ndest, server, client, kind="edit"):
        if type(var) == cdp.sms_variable:
            name = var.name
        else:
            name = var.name()
        vfrom = None
        vdest = None
        repeat = nfrom.get_repeat()
        print("############# %s" % repeat, nfrom.get_abs_node_path())
        if repeat and repeat.name() == name:
            vfrom = repeat
        else:
            for vfrom in nfrom.variables:
                if vfrom.name() == name:
                    break

        if type(ndest) == cdp.sms_node:
            vdest = ndest.variable
            while vdest:
                if vdest.name == name:
                    break
                vdest = vdest.next
            pnode = fullname(ndest)
        else:
            for vdest in ndest.variables:
                if vdest.name() == name:
                    break
            pnode = ndest.get_abs_node_path()
        pnode = nfrom.get_abs_node_path()

        if vfrom is None:
            return

        elif vfrom.name() != name:
            return

        if vdest is None:
            return

        if type(vdest) == cdp.sms_variable:
            vname = vdest.name
            vval = vdest.value
        else:
            vname = vdest.name()
            vval = vdest.value()
        if vname != name:
            return

        if vfrom.name() == vname and str(vfrom.value()) != str(vval) or FORCE:
            if type(vfrom) == Repeat:
                self.action(ndest.get_abs_node_path(), server, client,
                            kind="repeat", value=str(vfrom.value()),
                            msg="repeat requeue")
            else:
                self.action(pnode, server, client, kind=name,
                            value=str(vfrom.value()), msg="value update")

    def process(self, server=None, client=None, item=None, var=False):
        if server is None:
            self._check(self.sname, self.source, 1)

            for server in self.servers:
                server = server.replace('@', ':')
                if DEBUG:
                    print("#DBG: target is", server)
                host, port=server.split(':')

                if int(port) < 65536 and 0 == comm(
                        ECFLOWC + " --ping --port %s --host %s" % (port, host)):
                    client = ecflow.Client(server)
                    client.ch_register(False, self.register)
                    client.sync_local()
                    self._check(server, client)
                    for suite in self.register:
                        node = definitions[suite]
                        if type(node) == str:
                            node = self.source.get_defs().find_abs_node(node)
                        # try: node.add(Label("memo", "This suite was generated by mirror.py"))
                        # except: pass # used to sync, too late
                        # update variables
                        self.process(server, client, node, "repeat")
                        # self.process(server, client, node) # update statuses
                        # self.process(server, client, node, "edit") # update
                        # variables

                elif 0 == comm("SMS_PROG=%s %s %s" % (port, XPING, host)):
                    port = int(port)
                    os.putenv('SMS_PROG', "%d" % port)
                    os.environ['SMS_PROG'] = "%d" % port
                    cdp.cdp_init0()
                    passw = "1"
                    tout = 60
                    top = cdp.sms_cdp_client_login(
                        host, USER, passw, tout, port)
                    if top is None:
                        print("#ERR: node not found", host, port, USER, passw)
                        return

                    for suite in self.register:
                        node = definitions[suite]
                        if type(node) == str:
                            node = self.source.get_defs().find_abs_node(node)
                        # node.add(Label("memo", "This suite was generated by
                        # mirror.py"))
                        self.process(server, top, node, "repeat")
                        # self.process(server, top, node)
                        # self.process(server, top, node, "edit")

                else:
                    # child.report("abort","res")
                    print("server is not responding")

            if self.wait > 0:
                print("#MSG: start again... snap ", self.wait)
                time.sleep(self.wait)
                self.wait = 0
                self.source.sync_local()
                self.process()
            return

        elif server == self.sname:
            child.report("abort", "same name")  # DONT

        elif client is None:
            child.report("abort", "no client")  # DONT

        elif type(item) in (ecflow.Suite, ecf.Suite):

            if isinstance(client, cdp.sms_node):
                suite = cdp.sms_node_find(item.get_abs_node_path(), client)
                if suite:
                    if not sms_status[suite.status] == "suspended":
                        print("#ERR: top node shall be suspended - stop",
                              suite, server)
                        return
                    node = suite.kids
                    while node:
                        # print(node); print(node.name, server, client, var)
                        self.process(server, client, node, var)
                        node = node.next
            else:
                suite = client.get_defs().find_abs_node(item.get_abs_node_path())
                print(type(client), item.get_abs_node_path(), type(suite))
                if suite:
                    if not suite.is_suspended():
                        print("#ERR: top node shall be suspended - stop",
                              suite, server)
                        return
                    for node in suite.nodes:
                        self.process(server, client, node, var)
            return

        elif type(item) in (ecflow.Family, ecflow.Task):
            pnode = item.get_abs_node_path()
            nfrom = self.source.get_defs().find_abs_node(pnode)
            if isinstance(client, cdp.sms_node):
                ndest = cdp.sms_node_find(pnode, client)
            else:
                ndest = client.get_defs().find_abs_node(pnode)

            if nfrom is None:
                if var:
                    return  # ignore
                print("#ERR: node not found, please update mirror.py skel",
                      pnode, self.sname)
                child.report("abort", "nfrom")

            elif ndest is None:
                if var:
                    return  # ignore
                print("#ERR: node not found, replay?", pnode, server)
                return

            else:
                pass
            if nfrom:
                status = "%s" % nfrom.get_state()
            else:
                print(pnode, self.source)
                raise Exception

            if isinstance(ndest, cdp.sms_node):
                dstate = sms_status[ndest.status]
            else:
                dstate = "%s" % ndest.get_state()
            is_task = type(item) == ecf.Task

            for evs in nfrom.events:
                if isinstance(ndest, cdp.sms_node):
                    break  # TODO

                for evd in ndest.events:
                    if evd.name() == "" and evd.number() != evs.number():
                        continue
                    if evd.value() == evs.value():
                        break
                    print("#event", evs.name(), evd.number(), evs.number(),
                          evd.value(), evs.value())
                    if evs.value():
                        if 0:
                            client.alter(pnode, "change", "event",
                                         evs.name_or_number(), "set")
                        elif evs.name() == "":
                            client.alter(pnode, "change", "event",
                                         "%d" % evs.number(), "set")
                        else:
                            client.alter(pnode, "change", "event",
                                         evs.name(), "set")

            for evs in nfrom.meters:
                if isinstance(ndest, cdp.sms_node):
                    break  # TODO

                for evd in ndest.meters:
                    if evd.name() != evs.name():
                        continue
                    if evd.value() == evs.value():
                        break
                    print("#meter", evd.value(), evs.value())
                    try:
                        client.alter(pnode, "change", "meter",
                                     evs.name(), "%s" % evs.value())
                    except:
                        pass

            if status != dstate or FORCE:
                print("#WAR: diff ", pnode, status, dstate, )
                if status == "complete":
                    self.action(pnode, server, client, status, is_task,
                                msg="forced complete")
                    if not is_task:
                        return
                elif status == "queued":
                    self.action(pnode, server, client, status, is_task,
                                msg="forced queued")
                elif status in ("submitted", "active", "unknown", "aborted", ):
                    print("status is active/submit, ignored")
                else:
                    print(status)
                    raise Exception

            if type(item) in (ecflow.Suite, ecflow.Family):
                for node in item.nodes:
                    self.process(server, client, node, var)

            if var:
                for edit in item.variables:
                    self.update_var(edit, nfrom, ndest, server, client, var)

        elif type(item) == cdp.sms_node:
            pnode = fullname(item)
            nfrom = self.source.get_defs().find_abs_node(pnode)
            ndest = cdp.sms_node_find(pnode, client)
            if nfrom is None:
                if var:
                    return  # ignore
                print("#ERR: node not found, please update mirror.py skel",
                      pnode, self.sname)
                child.report("abort", "nfrom")

            elif ndest is None:
                if var:
                    return  # ignore
                print("#ERR: node not found, replay?", pnode, server)
                return

            else:
                pass
            if nfrom:
                status = "%s" % nfrom.get_state()
            else:
                status = "unknown"
            dstate = "%s" % x_status[ndest.status]
            is_task = x_type[item.type]

            if status != dstate or FORCE:
                print("#WAR: diff ", pnode, status, dstate, )
                if status == "complete":
                    self.action(pnode, server, client, status, is_task,
                                msg="forced complete")
                    if not is_task:
                        return

                elif status == "queued":
                    if is_task:
                        self.action(pnode, server, client, status, is_task,
                                    msg="forced queued")
                    else:
                        pass
                elif status in ("submitted", "active", "aborted", "unknown"):
                    print("status is active/submit, ignored")
                else:
                    print(status, dstate)
                    raise Exception

            if var:
                edit = item.variable
                while edit:
                    self.update_var(edit, nfrom, ndest, server, client, var)
                    edit = edit.next

            if x_type[item.type] == "task":
                return
            node = item.kids
            while node:
                self.process(server, client, node, var)
                node = node.next

        else:
            print(type(item))
            raise Exception

    def action(self, pnode, server, client, kind, is_task=False, value="20010101",
               msg=""):
        if not self.act:
            return

        self._protect_oper(server)

        print("#MSG:", msg)
        if type(client) == cdp.sms_node:
            host, port=server.split(':')
            CDP="/usr/local/apps/sms/bin/cdp -q -c "
            cmd=CDP + "'set SMS_PROG %s; login %s %s 1;" % (port, host, USER)
        if kind in ("active", "submitted", ):
            return
        elif FORCE:
            pass

        elif kind == "complete":
            if is_task:
                if DEBUG:
                    print("#DBG: force complete")
                if type(client) == cdp.sms_node:
                    cmd += "force complete %s; exit 0;'" % pnode
                    rc = comm(cmd)
                else:
                    client.force_state(pnode, ecflow.State.complete)
            else:
                if DEBUG:
                    print("#DBG: force rec complete")
                if type(client) == cdp.sms_node:
                    cmd += "force -r complete %s; exit 0;'" % pnode
                    rc = comm(cmd)
                else:
                    client.force_state_recursive(pnode, ecflow.State.complete)
                    client.sync_local()

        elif kind == "queued":  # and is_task:
            if DEBUG:
                print("#DBG: force queued")
            if type(client) == cdp.sms_node:
                cmd += "force queued %s; exit 0;'" % pnode
                rc = comm(cmd)
            else:
                client.force_state(pnode, ecflow.State.queued)

        elif kind == "repeat":
            raise Exception

        else:  # var
            if DEBUG:
                print("#DBG: update var", kind, value)
            if kind in ("queued", ):
                pass
            elif kind not in ("YMD", "JUL", "DAY", "SCHOST", "STHOST",
                              "YYYY", "MM", "INIBEGINDATE", "MIXTASK"):
                raise Exception(kind)
            if type(client) == cdp.sms_node:
                cmd0 = cmd
                if kind in ("YMD", ):
                    cmd += "requeue %s; " % (pnode)
                cmd += "alter -V %s:%s %s; exit 0;'" % (pnode, kind, value)
                rc = comm(cmd)
                cmd += "alter -l %s:%s %s; exit 0;'" % (pnode, kind, value)
                rc = comm(cmd)
            else:
                if kind in ("YMD", ):
                    client.requeue(pnode)
                client.alter(pnode, "change", "variable", kind, value)
                try:
                    client.alter(pnode, "change", "label", kind, value)
                except:
                    pass

        time.sleep(self.snap)

# ignore Family + queued while is might just be that we dont mirror all
# fam/tasks and inheritance get it already complete in dest while still
# queued in source


############################
def replay(path, defs=None):
    import ecf
    miss = "@undef@"
    host = os.getenv("ECF_HOST", miss)
    if host == miss:
        host = os.getenv("ECF_NODE", "localhost")
    port = os.getenv("ECF_PORT", 31415)
    client = ecf.Client(host, port)
    if defs is None:
        defs = gen_suite(host, port, path)
    if DEBUG:
        print(defs)
    print("#MSG: replacing %s in " % path, host, port, defs)
    client.replace(path, defs, 1, 1)
    return client


def comm(cmd, rem=1):
    import commands
    if rem and cmd in memo.keys():
        return memo[cmd]
    (rc, res) = commands.getstatusoutput(cmd)
    print("#COMM:", rc, res, cmd)
    if rem:
        memo[cmd] = rc
    return rc


def sms_replay(path, defs=None):
    import commands
    import ecf
    miss = "@undef@"
    host = os.getenv("ECF_HOST", miss)
    if host == miss:
        host = os.getenv("SMSNODE", "localhost")
    prog = os.getenv("SMS_PROG", 314159)
    deff = "mirror.tmp"
    sname = path.split('/')[1]
    if defs is None:
        defs = Defs()
        if type(definitions[sname]) == str:
            server = "localhost:31415"
            client = ecflow.Client(server)
            client.ch_register(False, [sname, ])
            client.sync_local()
            defs = client.get_defs()
        else:
            defs.add_suite(definitions[sname])
        if DEBUG:
            print(defs)
        fd = open(deff, "w")
        print(defs, file=fd)
        fd.close()
    cmd = "cdp -c 'set SMS_PROG %s; login %s %s 1; play -r %s %s; exit 0'" % (
        prog, host, USER, path, deff)
    rc = comm(cmd)


############################
class TestMirror(unittest.TestCase):
    """ a test case """

    def test_1(self, test_ok=1): replay('/' + sname, gen_suite())


memo = dict()
############################
if __name__ == '__main__':
    try:
        OPTS, ARGS=getopt.getopt(
            sys.argv[1:], "fhi:m:n:p:rstw",
            ["force", "help", "interval", "mirror", "number", "path", "replay",
             "sms", "test", "wait", ])
    except getopt.GetoptError as err:
        print("#what?", usage())

    global FORCE
    FORCE = False  # FORCE = True
    MIRROR = None
    WAIT = False
    PATH = None
    NUM = None
    INTERVAL = 30
    REPLAY = False
    ECFLOW = True

    for o, a in OPTS:
        if o in ("-h", "--help"):
            usage()
        elif o in ("-f", "--force"):
            FORCE = True
        elif o in ("-i", "--interval"):
            INTERVAL = int(a)
        elif o in ("-m", "--mirror"):
            MIRROR = a
        elif o in ("-n", "--number"):
            NUM = int(a)
        elif o in ("-p", "--path"):
            PATH = a
        elif o in ("-r", "--replay"):
            REPLAY = True
        elif o in ("-s", "--sms"):
            ECFLOW = False
        elif o in ("-t", "--test"):
            unittest.main(argv=[sys.argv[0]])
            sys.exit(0)
        elif o in ("-w", "--wait"):
            WAIT = True

    child = Child()
    if not MICRO[0] in "$ECF_PORT$":
        MIRROR = "$DESTINATIONS$"
        PATH = "/$FAMILY1$"
        print("#MSG: ssh %s kill -9 %d" % (os.uname()[1], os.getpid()))

    if PATH and REPLAY:
        print("#MSG: path and replay", PATH, REPLAY)
        if ECFLOW:
            if 0:
                replay(PATH)
            else:
                s = PATH.split('/')[1]
                if not "mirror" in PATH:
                    defs = Defs()
                    defs.add_suite(definitions[s])
                else:
                    defs = None
                replay(PATH, defs)
        else:
            sms_replay(PATH)

    elif MIRROR and PATH:
        mirror = Mirror(MIRROR, PATH, REPLAY)
        print("#MSG:step")
        mirror.process()
        print("#MSG:step")
        child.report("complete")
        print("#MSG:step")

    elif NUM and PATH:
        check3of5(NUM, PATH, WAIT, INTERVAL)

    else:
        usage()
    print("#MSG:step")

    child.report("complete")
"""
$nopp
$end
"""



#pyflow Method

#pyflow is a pure python module on top of ecflow native API ;
# it is proposed to facilitate suites definitions. 
# It is not yet provided in default standard location. 
# It can be used, adding this one line to the definition script.

import sys; sys.path.append("/home/ma/emos/def/pyflow")

# Enter the following python code into a file i.e. test.py :

#!/usr/bin/env python2.7
import os
from pyflow import (Suite, Family, Task, Variable)
   
print "Creating suite definition"   
with Suite("test" ) as suite: 
 ECF_HOME = os.path.join(os.getenv("HOME"),  "course") # python vriable
 Variable("ECF_HOME", ECF_HOME)  # suite variable
 Task("t1")
print suite.ecflow_definition() 

#Then run as a python script:


#pyflow
#Using pyflow, header files are generated and placed in the expected location (ECF_INCLUDE) when the suite is deployed:
suite.deploy_suite()


#!/usr/bin/env python
from distutils.core import setup, Extension
import os
import sys
import glob

# ========================================================================
# Usage:
#   cd $WK; cd ecbuild/release/; sh -x ../../cmake.sh release; make ; make install
#   calling 'sh -x ../../cmake.sh release' will generate setup.py from setup.py.in
# Issues:
#   Can not test cmake install, since it will always install to 
#     /usr/local/apps/python/current/lib/python2.7/site-packages/ecflow
#
# To test install manually, we can:
#   cd $WK/Pyext
#   rm -rf build/   # for a clean build
#   python setup.py build_ext
#   python setup.py install --home=~
#
# See: http://docs.python.org/2/distutils/apiref.html?highlight=extension#distutils.core.Extension
#
# ========================================================================
# AIX: On AIX we need custom compile and link flags:
# The final link line is wrong: xlC_r xlc_r  , 
#                              /usr/local/lpp/vacpp11109/usr/vacpp/bin/.orig/xlC_r: 1501-228 (W) input file xlc_r not found
#           
# Removing the duplicated xlc_r, We then get a little further,
#         ld: 0711-317 ERROR: Undefined symbol: .main
#      Using -bnoquiet
#        .main                     [12]    ER PR crt0_64.s(/lib/crt0_64.o)
#                                          00000070 .text    R_RBR    [34]    .__start
# This is because it is missing -G (generate dynamic library)
# When we manually add -G, we sucessfully link(ecflow.so)
#
#      Both of these can be fixed, by overriding the Link line with:
#      export LDSHARED="xlC_r -G -Wl,-bI:/usr/local/apps/python/2.7.2-01/lib/python2.7/config/python.exp"
#
# To find out what python is going to use:
#         python -c "from distutils import sysconfig; print sysconfig.get_config_vars('LDSHARED')[0]"
# On AIX, it does not use this  value ? 
#
#      See python issue 18235  _sysconfigdata.py wrong on AIX installations
# ==========================================================================

# ==========================================================================
# Permissions: The permission of congfigured file are wrong: 2 choices
# o Make sure origin file, has the right permissions
# o Copy file to different directory, and change the permissions
#   since file(COPY) does rename files
#   configure_file(setup.py.in /tmp/ma0/workspace/bdir/release/ecflow/CMakeFiles/setup.py)
#       now copy the temporary into the final destination, setting the permissions
#   file(COPY /tmp/ma0/workspace/bdir/release/ecflow/CMakeFiles/setup.py
#     DESTINATION /tmp/ma0/workspace/bdir/release/ecflow
#     FILE_PERMISSIONS OWNER_READ OWNER_WRITE OWNER_EXECUTE GROUP_READ
#     GROUP_EXECUTE WORLD_READ WORLD_EXECUTE)
# ============================================================================
        
#=============================================================================
# define the directories to search for include files
# to get this to work, you will need to include the path
# to your boost installation and  ecflow includes
boost_root=os.getenv("BOOST_ROOT") 
include_dirs = [ "/tmp/ma0/workspace/ecflow/Pyext/../ACore/src", 
                 "/tmp/ma0/workspace/ecflow/Pyext/../ANattr/src", 
                 "/tmp/ma0/workspace/ecflow/Pyext/../ANode/src",
                 "/tmp/ma0/workspace/ecflow/Pyext/../Base/src",
                 "/tmp/ma0/workspace/ecflow/Pyext/../Base/src/cts",
                 "/tmp/ma0/workspace/ecflow/Pyext/../Base/src/stc",
                 "/tmp/ma0/workspace/ecflow/Pyext/../CSim/src", 
                 "/tmp/ma0/workspace/ecflow/Pyext/../Client/src",
                 "/tmp/ma0/workspace/ecflow/Pyext/src",
                 boost_root,
               ]

# define the library directories to include any extra libraries that may be needed.
# Give preference to release libs   
boost_lib_dir = boost_root + "/stage/lib/"
library_dirs = ['/tmp/ma0/workspace/bdir/release/ecflow/ACore',
                '/tmp/ma0/workspace/bdir/release/ecflow/ANattr/',
                '/tmp/ma0/workspace/bdir/release/ecflow/ANode/',
                '/tmp/ma0/workspace/bdir/release/ecflow/Base/', 
                '/tmp/ma0/workspace/bdir/release/ecflow/CSim/', 
                '/tmp/ma0/workspace/bdir/release/ecflow/Client/', 
                boost_lib_dir  
               ]

# define the libraries to link with this includes the boost lib
libraries = [ 'core' , 'nodeattr', 'node', 'base', 'libsimu', 'libclient',
              'boost_system-mt',
              'boost_serialization-mt',
              'boost_filesystem-mt',
              'boost_program_options-mt',
              'boost_date_time-mt', 
              'boost_python-mt' ]

# Using gcc-4.8 with boost 1.53 on linux requires these flags
extra_compile_args = ['-ftemplate-depth-512', '-Wno-unused-local-typedefs' ]
extra_link_args = []

# extra compile flags needed for AIX only
# Note setup.py will add -q64 -qcpluscmt -DNDEBUG  automatically
# Note: two extra_compile_args, debug and release, use the debug for testing and faster compiles
if sys.platform.startswith("aix"):
   extra_compile_args =  [ '-qsuppress=1540-0198', '-O3', '-qstrict', '-qfuncsect', '-qeh', '-qrtti'  ]
   #extra_compile_args = [ '-qsuppress=1540-0198',  '-qNOOPTimize', '-qnoinline', '-qfullpath', '-qfuncsect', '-qeh', '-qrtti' ]
   extra_link_args= [ '-lpthread', '-ldl', '-bbigtoc' ]  # '-G', '-noipath', 
 

# create the extension and add it to the python distribution
# o glob.glob(os.path.join('src', '*.cpp'))
#   This expand the list of cpp files that need to be compiled
#
# The configuation below installs to:
#   o /usr/local/apps/python/current/lib/python2.7/site-packages/ecflow
#     when, "python setup.py install" is used:
#   lib/python2.7/site-packages/ecflow/ecflow.so
#                                     __init__.py               
setup( name='ecflow', 
       version='4.8.0', 
       author      = 'ECMWF',
       description = """ecflow Python interface""",
       packages = [ 'ecflow' ],
       package_dir={'ecflow': '/tmp/ma0/workspace/ecflow/Pyext/ecflow'},
       ext_modules=[ Extension( 
                              'ecflow.ecflow', 
                              glob.glob(os.path.join('/tmp/ma0/workspace/ecflow/Pyext/src', '*.cpp')), 
                              include_dirs=include_dirs, 
                              library_dirs=library_dirs, 
                              libraries=libraries,
                              extra_compile_args=extra_compile_args,
                              extra_link_args=extra_link_args
                              )
                    ],
   )



#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009-2017 ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

# This test checks trigger evaluation uses date arithmetic when
# the expression reference a repeat DATE variable.
# Addition and subtraction for repeat date, should follow data arithmetic
# The repeat variable in the expression must be on the LHS(left hand side)
# of the expression, otherwise integer arithmetic is used
import ecflow
 
if __name__ == "__main__":
    print("####################################################################")
    print("Running ecflow version " + ecflow.Client().version() + " debug build(" + str(ecflow.debug_build()) + ")")
    print("####################################################################")
 
    defs = ecflow.Defs()
    s1 = defs.add_suite("s1");
    t1 = s1.add_task("t1").add_repeat( ecflow.RepeatDate("YMD",20090101,20091231,1) );
    t2 = s1.add_task("t2").add_trigger("t1:YMD ge 20100601");
 
    # Check trigger expressions
    assert len(defs.check()) == 0,  "Expected no errors in parsing expressions."
 
    # Initial value of repeat is 20090101 hence trigger should fail to evaluate
    assert t2.evaluate_trigger() == False, "Expected trigger to evaluate. 20090101 >= 20100601"
   
    # Check end of month - 1
    t2.change_trigger("t1:YMD - 1 eq 20081231")
    assert t2.evaluate_trigger(), "Expected trigger to evaluate. 20090101 - 1  == 20081231"

    # check addition
    t2.change_trigger("t1:YMD + 1 eq 20090102");    
    assert t2.evaluate_trigger(), "Expected trigger to evaluate. 20090101 + 1  == 20090102"

    # Check the end of each month + 1
    t1.delete_repeat();
    t1.add_repeat( ecflow.RepeatDate("YMD",20090131,20101231,1)) 
    t2.change_trigger("t1:YMD + 1 eq 20090201");   
    assert t2.evaluate_trigger(), "Expected trigger to evaluate. 20090131 + 1  == 20090201"

    print("All Tests pass")
    



#!/usr/bin/env python
# coding=utf-8
# This software is provided under the ECMWF standard software license agreement.
""" a layer over raw ecflow api

use 'export ECF_DEBUG_LEVEL=10' to remove warning message related to
variables overwrite

specialisation module for ecFlow nodes/attributes:
initially it was simple derivation
then it is composition

"""
from __future__ import print_function    # (at top of module)
import sys
import pwd
import os
import unittest
# try:
#     from ecflow import TimeSlot
#     from ecflow import JobCreationCtrl as JobCreationCtrl
#     from ecflow import ChildCmdType
#     from ecflow import ZombieType, ZombieAttr, ZombieUserActionType
# except:
#     print "# ecf.py cannot import few types"
try:
    import ecflow
except:
    loc = "/usr/local/apps/ecflow/current/lib/python2.7/site-packages/ecflow"
    sys.path.append(loc)
    loc = "/usr/local/lib/python2.7/site-packages"
    sys.path.append(loc)
    import ecflow
# import ecflow; 
ecflow.Ecf.set_debug_level(3)

DEBUG = 0
DECORATE = "ALL"
# DECORATE = "ONLY_TRIGGER"
# DECORATE = "NO_TRIGGER"
# DECORATE = "ONLY_EVENT"
# DECORATE = "NO_ATTRIBUTE"

USE_TIME = True
USE_LATE = False
USE_TRIGGER = True
USE_LIMIT = True
CPUTIME = True

if DECORATE == "NO_TRIGGER":
    USE_TRIGGER = False
    USE_LIMIT = True
    USE_EVENT = True
elif DECORATE == "ALL":
    USE_TRIGGER = True
    USE_LIMIT = True
    USE_EVENT = True
else: raise Exception


def get_username():
    return pwd.getpwuid(os.getuid())[0]


def get_uid():
    return pwd.getpwnam(get_username()).pw_uid

def obsolete():
    if 1: return
    if 1: raise Exception

def raise_int(arg):
    if 0: return
    if "/none" in arg: raise Exception("None!!!", arg)
    if arg in ("1==0", "0==1", "1==1", "0==0"): return
    if arg[0] in ("0", "1", "2", "3", "4", "5", "6", "7", "8", "9"):
        raise Exception("please add ./", arg)

def set_raw_mode(kind=True):
    import sms2ecf
    sms2ecf.RAW_MODE = kind


LINKTASK = False  # FD
# LINKTASK = True  # RD mode v we turn task into families,
                   # with a leaf a task matching script name
def linktask(name, real="get", add=None, fam=None, task=True):
    """ give a change to avoid UNIX links time consuming for maintenance, update
    perforce git
    name: original name
    real: script/task wrapper name
    add: attributes that shall be added to the leaf task, eg Label, Event, Meter
    """
    if fam is None: fam = name
    try:
        import parameters as ip
        # if ip.USER == "rdx": global LINKTASK; LINKTASK = True  # acceptance
        # if ip.USER == "emos": global LINKTASK; LINKTASK = False  # acceptance
    except: pass
    if LINKTASK: return Task(name)  # RD???
    return Family(fam).add(
        Task(real).add(add, If(task, Edit("TASK", name))))


def translate(name, value=None):
    """ Translate from sms to ecflow """
    try: import sms2ecf
    except: return name, value
    def is_sms():
        if sys.argv[-1] in ("sms", "od3", "ode", "map", "od", "od2", "ligarius"):
            sms2ecf.ECF_MODE = "sms"
            return True
        return sms2ecf.ECF_MODE == "sms"
    if is_sms():
        sms2ecf.ECF_MODE = "sms"
        return sms2ecf.translate(name, value)
    return name, value


DEFS = None  # general case when not in parameters.py
# USE_EXTERN = 1
# NO_EXTERN_ALONE = 0 # set to 1 for test/activate
class Extern(object):
    """ extern may be collected and added as soon as Defs is created """

    def __init__(self, path):
        if type(path) in (tuple, list, str): pass
        else: raise Exception

        try:
            import parameters as ip
            if type(path) in (tuple, list):
                for item in path: Extern(item)
            elif type(path) in (str, ):
                ip.DEFS.load.add_extern(path)
            return
        except:
            global DEFS
            if DEFS is None: DEFS = Defs()

        if type(path) == tuple or type(path) == list:
            for ppp in path: Extern(ppp)

        elif type(path) == str:
            if DEBUG: print("#MSG: extern", path)
            if ":" in path: Extern(path.split(":")[0])
            elif DEFS: DEFS.load.add_extern(path)

        elif path is None: pass

        else:
            raise Exception(type(path), path)

    def add_to(self, node): return None


class State(object):
    """ this class aims at affording a user the possibility to add Triggers as
    t1 = Task("t1")
    Task("ts").add(Trigger(t1 == COMPLETE))

    SUBMITTED, ACTIVE, SUSPENDED, ABORTED, QUEUED, COMPLETE, UNKNOWN
    are instance of this class.
    """

    def __init__(self, state):
        """ store the status """
        self.state = str(state)

    def __str__(self):
        """ translate into string """
        return "%s" % self.state

    def __eq__(self, arg):
        """ when == is used,
        we should care about task name starting with 0-9"""
        if type(arg) == str:
            if type(arg[0]) == int:
                add = "./"
            else: add = ""
            return add + arg + " == " + self.state
        elif isinstance(arg, Node):
            return arg.fullname() + " == " + self.state
        elif isinstance(arg, ecflow.Node):
            return arg.get_abs_node_path() + " == " + self.state
        return False

    def __ne__(self, arg):
        """ aka != """
        if type(arg) == str:
            if type(arg[0]) == int:
                add = "./"
            else: add = ""
            return add + arg + " != " + self.state
        elif isinstance(arg.load, ecflow.Node):
            return arg.load.get_abs_node_path() + " != " + self.state
        return False

    def value(self):
        """ return state """
        return self.state

    def eval(self, node):
        """ return state """
        return self.state == node.get_state()

SUBMITTED = State("submitted")
ACTIVE = State("active")
SUSPENDED = State("suspended")
ABORTED = State("aborted")
QUEUED = State("queued")
COMPLETE = State("complete")
UNKNOWN = State("unknown")


class Attribute(object):
    """ generic attribute to be attached to a node"""

    def add_to(self, node):
        """ use polymorphism to attach attribute to a node"""
        raise Exception("ERR: virtual class")  # return node


class Label(Attribute):
    """wrap around label"""

    def __init__(self, name, msg=""):
        self.load = ecflow.Label(name, msg)

    def add_to(self, node):
        """add_label"""
        node.load.add_label(self.load)
        return node

class Meter(Attribute):
    """wrap around meter"""

    def __init__(self, name, beg, end, thr=None):
        if thr is None: thr = end
        self.load = ecflow.Meter(name, beg, end, thr)

    def add_to(self, node):
        """ add_meter"""
        node.load.add_meter(self.load)
        return node


class Event(Attribute):
    """ wrap around event"""

    def __init__(self, name):
        self.load = ecflow.Event(name)

    def add_to(self, node):
        """ add_event"""
        node.load.add_event(self.load)
        return node


class Inlimit(Attribute):
    """ a class to host a path for a limit
        silently ignore if USE_LIMIT is False,
        (in debug mode) """

    def __init__(self, fullpath):
        self.data = None
        if USE_LIMIT:
            try:
                path, name = fullpath.split(":")
            except:
                name = fullpath
                path = ""
            if name is None:
                raise Exception
            if '-prod' in name: name = name.split('-')[0]  # vdiss-prod
            if '-' in name: name = name.replace('-', '_')
            if "None" in path or "None" in name:
                raise Exception(path, name, fullpath)
            self.data = ecflow.InLimit(name, path)
            self.path_ = path
            self.name_ = name

    def add_to(self, node):
        """ add_inlimit"""
        if not USE_LIMIT:
            return
        if self.data is None:
            raise Exception
        node.load.add_inlimit(self.data)
        return node

    def value(self):
        """ get limit fullpath-name """
        return self.path_ + ":" + self.name_

    def name(self):
        """ get limit name """
        return self.name_


class InLimit(Inlimit): pass


def item_to_string(prev, index, name):
    """ index not used... """
    if name is None: return ""
    if name is "": return prev
    if "<main>" in name: raise Exception
    if "==" in name: return name
    if len(name) > 0:
        if name[0].isdigit():
            pre = "./"  # "0123456789":
            name = pre + name
    if prev is None or prev == "":
        if ':' in name or name == "":
            return name
        else: return "%s == complete" % name
    if ' or ' in name:
        if "complete" in name:
            return prev + "%s" % name
        elif ":" in name:
            return prev + "%s" % name
        else:
            return prev + "%s == complete" % name
    elif ':' in name:  # event: and x:1
        return prev + " and %s" % name
    elif 'complete' in name:  # complete: and x == complete
        return prev + " and %s" % name
    else:
        return prev + " and %s == complete" % name


class Trigger(Attribute):
    """ add trigger (string, list of task names, or directly
           expression and: and'ed (True) or or'ed (False) unk: add or
           [name]==unknown for RD """

    def __init__(self, expr, unk=False, anded=True):
        self.expr = ""
        if expr is None or expr == "":
            return
        if type(expr) == str:
            if "%s" in expr: raise Exception("ERR:", expr)
            if "%d" in expr: raise Exception("ERR:", expr)
            self.expr = expr
            try:
              import parameters as ip
              dim = len(ip.SELECTION)
              for ploc in expr.split():
                if ploc[0] == '/' and ip.SELECTION != ploc[1:dim + 1]:
                    if ploc in ("eq", "ne", "==", "!=", "and", "or",
                                "active", "complete", "queued", ):
                        continue
                    if ploc.isdigit():
                        continue
                    Extern(ploc.replace("==complete", "").replace(")", ""))
            except: pass
            # NO_EXTERN_ALONE = 0 # set to 1 for test/activate
            return
        if type(expr) == tuple:
            prep = list(expr)
            expr = prep

        if type(expr) == Task:
            pnode = expr.fullname()
            if pnode[0] != '/': pnode = expr.name()
            try:
                if ic.psel() not in pnode: pnode = expr.name()
            except: pass
            # if 1: raise Exception("really?", pnode)
            self.expr += item_to_string(None, 0, pnode)

        elif type(expr) in (list, set, tuple):
            for index, name in enumerate(expr):
                if name is None:
                    continue
                pre = ""

                if name is None: continue
                if type(name) in (set, list, tuple):
                    print("#WAR: please avoid list in list", expr, name)
                    if 1: print("#WAR: list in list")
                    elif 1: raise Exception("please avoid list in list", expr, name)
                    name = ""
                    for index, another in enumerate(name):
                        name += item_to_string(name, index, another)
                elif type(name) in (Node, Task, Family, Suite):
                    if DEBUG: print(name.name())
                    fullname = name.fullname()
                    if fullname[0] == '/':
                        try:
                            import inc_common as ic
                            if ic.psel() not in fullname:
                                fullname = name.name()
                        except: print("#WAR: trigger expcept")
                    name = fullname
                elif type(name) in (ecflow.Task, ecflow.Family):
                    name = name.name()
                elif type(name) == str:
                    pass
                else:
                    raise Exception(type(name), name)

                if "==" in name: pass
                elif len(name) > 0:
                    # if "==" in name: pass
                    # if "1==0" in name or "0==1" in name or "1==1" in name: pass
                    if name[0].isdigit(): pre = "./"  # "0123456789":

                self.expr = item_to_string(self.expr, index, name); continue ##

        elif type(expr) in (ecflow.Expression,
                            ecflow.PartExpression):
            self.expr = ecflow.Expression(str(item))

        elif type(expr) in (Family, Task):
            fullname = expr.fullname()
            if fullname[0] == '/':
                try:
                    import inc_common as ic
                    if ic.psel() not in fullname:
                        fullname = name.name()
                except: print("#WAR: trigger expcept")
            self.expr = fullname
            # check full name is really full name, not /name
            # print("#REM:", self.expr)
            if self.expr == "/%s" % expr.name(): self.expr = expr.name()
        else:
            raise Exception("what? trigger?", type(expr))

        if "YMD+1" in self.expr: raise Exception(self.expr)
        # NO_EXTERN_ALONE = 0 # set to 1 for test/activate

    def add_to(self, node):
        if self.expr is None or self.expr == "" or not USE_TRIGGER:
            return None
        if 0 and "/make/setup" in "%s" % self.expr:  # help DEBUG
            raise Exception(self.expr, node.fullname())
        if node.load.get_trigger() is None:
            node.load.add_trigger(self.expr)
        else:
            node.load.add_part_trigger(ecflow.PartExpression(self.expr, True))
        return node

class TriggerImpossible(Trigger):
    """attribute to be added to node when it is not expected to run any task"""

    def __init__(self):
        """ add an 'impossible trigger', for a task not to run """
        super(TriggerImpossible, self).__init__("1==0")


class TriggerAlways(Trigger):
    """attribute to be added to node when it is not expected to run any task"""

    def __init__(self):
        """ add an 'impossible trigger', for a task not to run """
        super(TriggerAlways, self).__init__("1==1")


class Complete(Trigger):
    """ class to host complete expression, added later to a node"""

    def __init__(self, expression, unk=False, anded=False):
        super(Complete, self).__init__(expression, unk, anded)

    def add_to(self, node):
        if USE_TRIGGER and self.expr is not None:
            if "%s" in self.expr: raise Exception
            node.load.add_complete(self.expr)
        return node


class CompleteAlways(Complete):
    """ attribute """

    def __init__(self):
        """ always True as soon as evaluated """
        super(CompleteAlways, self).__init__("1==1")


class Clock(Attribute):
    """ wrapper to add clock """

    def __init__(self, arg="24:00", hybrid=0):
        if type(arg) == str:
            hybrid = "hybrid" in arg
            hhh, mmm, sss = [0, 0, 0]
            try:
                hhh, mmm, sss = arg.split(':')
                self.data = ecflow.Clock(hhh, mmm, sss, hybrid)
            except:
                self.data = ecflow.Clock(hybrid)
        else:
            self.data = ecflow.Clock(arg)

    def add_to(self, node):
        if type(node) != Suite:
            print("#WAR: clock can only be attached to suite node,\n",
                  "#WAR: clock is ignored")
            return
        node.load.add_clock(self.data)
        return node


class Autocancel(Attribute):
    """ wrapper to add time """

    def __init__(self, arg):
        if type(arg) == str:
            if ':' in arg:  # hh:mm +hh:mm
                hhh, mmm = arg.split(':')
                rel = '+' in arg
                self.data = ecflow.Autocancel(int(hhh), int(mmm), rel)
            else:
                self.data = ecflow.Autocancel(int(arg))
        else:
            self.data = ecflow.Autocancel(arg)  # days

    def add_to(self, node):
        node.load.add_autocancel(self.data)
        return node


class Autocancel(Autocancel): pass


class Script(Attribute):
    def __init__(self, script):
      self._script = script

    def add_to(self, node):
        print("#WAR: Script attribute, not yet ... ignored")
        return node


class Time(Attribute):
    """ wrapper to add time """

    def __init__(self, arg):
        self.data = arg

    def add_to(self, node):
        if USE_TIME and self.data is not None:
            node.load.add_time(self.data)
        return node

class Today(Time):
    """ wrapper to add time """

    def add_to(self, node):
        if USE_TIME and self.data is not None:
            node.load.add_today(self.data)
        return node


class Cron(Time):
    """ wrapper to add time """

    def __init__(self, bes, wdays=None, days=None, months=None):
        import argparse
        self.data = ecflow.Cron()

        if not ("-w" in bes or "-m" in bes or "-d" in bes):
            self.data.set_time_series(bes)
            return

        parser = argparse.ArgumentParser()
        parser.add_argument("-w", nargs='?', default=0, help="weekdays")
        parser.add_argument("-d", nargs='?', default=0, help="days")
        parser.add_argument("-m", nargs='?', default=0, help="months")
        parser.add_argument("arg", type=str, help="begin end step")
        parsed = parser.parse_args(bes.split())

        if parsed.w:
            self.data.set_week_days([int(x) for x in parsed.w.split(',')])
        if parsed.d:
            self.data.set_week_days([int(x) for x in parsed.d.split(',')])
        if parsed.m:
            self.data.set_months([int(x) for x in parsed.m.split(',')])
        self.data.set_time_series(parsed.arg)

    def add_to(self, node):
        if USE_TIME and self.data is not None:
            node.load.add_cron(self.data)
        else: print("#WAR: ignoring: %s" % self.data)
        return node


class Date(Time):
    """ wrapper to add date """

    def __init__(self, arg, mask=False):
        super(Date, self).__init__(arg)
        self.mask = mask

    def add_to(self, node):
        if USE_TIME and self.data is not None:
            # ??? emos avoids dates, datasvc would not
            if self.mask:
                node.load.add_variable("DATEMASK", self.data)
            else:
                ddd, mmm, yyy = self.data.split('.')
                if ddd == '*':
                    ddd = 0
                if mmm == '*':
                    mmm = 0
                if yyy == '*':
                    yyy = 0

                node.load.add_date(int(ddd), int(mmm), int(yyy))
                # node.add_date(self.data)
        return node


class Day(Date):
    """ wrapper to add day """

    def add_to(self, node):
        if not USE_TIME or self.data is None:
            return
        if isinstance(self.data, str):
            days = {"monday": ecflow.Days.monday,
                    "sunday": ecflow.Days.sunday,
                    "tuesday": ecflow.Days.tuesday,
                    "wednesday": ecflow.Days.wednesday,
                    "thursday": ecflow.Days.thursday,
                    "saturday": ecflow.Days.saturday,
                    "friday": ecflow.Days.friday,

                    "mon": ecflow.Days.monday,
                    "sun": ecflow.Days.sunday,
                    "tue": ecflow.Days.tuesday,
                    "wed": ecflow.Days.wednesday,
                    "thu": ecflow.Days.thursday,
                    "sat": ecflow.Days.saturday,
                    "fri": ecflow.Days.friday, }
            node.load.add_day(ecflow.Days(days[self.data]))
        else:
            node.load.add_day(ecflow.Days(self.data))
        return node


class Defstatus(Attribute):
    """ add defstatus attribute"""

    def __init__(self, kind):
        if type(kind) == str:
            kinds = {"suspended": ecflow.DState.suspended,
                     "aborted": ecflow.DState.aborted,
                     "complete": ecflow.DState.complete,
                     "active": ecflow.DState.active,
                     "submitted": ecflow.DState.submitted,
                     "unknown": ecflow.DState.unknown,
                     "queued": ecflow.DState.queued, }
            self.data = kinds[kind]
        else: self.data = kind

    def add_to(self, node):
        node.load.add_defstatus(self.data)
        return node


class Defcomplete(Defstatus):
    """ wrapper to add defstatus complete """

    def __init__(self):
        pass

    def add_to(self, node):
        node.add(Defstatus("complete"))
        return node


class DefcompleteIf(Defcomplete):
    """ wrapper to add conditional defstatus complete
    just change name to make it explicit """

    def __init__(self, arg=True):
        self.data = arg

    def add_to(self, node):
        if self.data:
            node.add(Defstatus("complete"))
        # else: node.defstatus("queued") # in comment to prevent
        # overwrite when using multiple defcomplete
        return node


class Limit(Attribute):
    """ wrapper to add limit """

    def __init__(self, name=None, size=1, inlimit=0):
        self.name = name
        self.size = size
        self.addi = inlimit

    def add_to(self, node):
        if USE_LIMIT and self.name is not None:
            if type(self.name) is dict:
                if LINKTASK: items = self.name.items()   # 45r1 COMPARE
                else: items = sorted(self.name.items())  # EMOS prefers
                for name, size in items:
                # for name in sorted(self.name.keys()):
                    size = self.name[name]
                    node.load.add_limit(name, size)
                    if self.addi:
                        node.load.add_inlimit(name)
                    else:
                        pass
            else:
                node.load.add_limit(self.name, self.size)
                if self.addi: node.load.add_inlimit(name)
        return node

# class Limits(Limit): pass

class Late(Attribute):
    """ wrapper around late, to be add'ed to families and tasks """

    def __init__(self, arg):
        self.data = None
        if not USE_LATE:
            # print("#MSG: late is disabled")
            return
        sub = False
        act = False
        com = False
        rel = False
        self.data = ecflow.Late()
        for item in arg.split(" "):
            if item == "-s":
                sub = True
            elif item == "-c":
                com = True
            elif item == "-a":
                act = True
            else:
                hour, mins = item.split(":")
                rel = "+" in hour
                if "+" in hour:
                    hour = hour[1:]
                if sub:
                    self._add_sub(hour, mins)
                elif com:
                    self._add_com(hour, mins, rel)
                elif act:
                    self._add_act(hour, mins)
                sub = False
                act = False
                com = False

    def _add_sub(self, hour, mins):
        """ submitted"""
        self.data.submitted(ecflow.TimeSlot(int(hour), int(mins)))

    def _add_com(self, hour, mins, rel):
        """ complete"""
        self.data.complete(ecflow.TimeSlot(int(hour), int(mins)), rel)

    def _add_act(self, hour, mins):
        """ active"""
        self.data.active(ecflow.TimeSlot(int(hour), int(mins)))

    def add_to(self, node):
        if USE_LATE and self.data is not None:
            try: node.load.add_late(self.data)
            except:
                if DEBUG: print("#WAR: late is already attached")
        return node


def python_true(key, val):
    if "%s" % val == "True":
        if key == "HYPERTHREADING": return  # OK
        print("WAR: really???", key, val)


class Edit(Attribute):
    """ dedicated class to enable variable addition with different
    syntax """

    def _set_tvar(self, key, val):
        """ facilitate to load a ecflow suite to SMS, translating
        variable names"""
        keyt, edit = translate(str(key), str(val))
        # if keyt == "SCHOST" and val == "ccb": raise Exception
        if self.data is None:
            self.data = ecflow.Variable(keyt, edit)
        else:
            next = self.next
            self.next = Edit(keyt, edit, next)

    def __init__(self, __a=None, __b=None, __next=None, *args, **kwargs):
        self.data = None
        self.next = __next

        if len(args) > 0:
            if type(args) == list:
                for item in args.iteritems():
                    key = item.name()
                    val = item.value()
                    python_true(key, val)
                    self._set_tvar(item.name(), item.value())
            elif type(args) == tuple:
                for key, val in args.items():
                    python_true(key, val)
                    self._set_tvar(key, val)
            else:
                raise Exception
        if len(kwargs) > 0:
            # for key, val in kwargs.items():
            # for key in sorted(kwargs.keys()):  # FIXME
            for key in kwargs.keys():
                val = kwargs[key]
                python_true(key, val)
                self._set_tvar(key, val)
        if type(__a) == dict:
            # for key, val in __a.items():
            # for key in sorted(__a.keys()):
            for key in __a.keys():  # FIXME
                val = __a[key]
                python_true(key, val)
                self._set_tvar(key, val)
        elif type(__a) == tuple:
            raise Exception
        elif type(__a) == list:
            raise Exception
        elif type(__a) == Variable:
            self.data = __a
        elif __a is not None and __b is not None:
            python_true(__a, __b)
            self._set_tvar(__a, __b)
        elif __a is None and __b is None:
            pass
        else:
            raise Exception(__a, __b, __next, args, kwargs)

    def add_to(self, node):
        if self.data is not None:
            edit = "%s" % self.data
            if "CPUTIME" in edit and not CPUTIME: pass
            else: node.load.add_variable(self.data)
            if node.fullname() == "/o" and "QUEUE 'emos" in edit:
                raise Exception(node.fullname(), edit)
            if 1:  # try:  # Operators' request
                labels = {"WSHOST": "infopws",
                          "SCHOST":"infopsc",
                          "HOST": "infophs", }
                # if "seas" in ip.SELECTION: pass
                for key in labels.keys():
                  try:
                    if "ECF_JOB_CMD" in edit:
                        find = "%" + key + "%"
                        if find in edit: node.add_label("infopcmd", key)
                    else:
                        find = "edit %s " % key
                        if find in edit: node.add_label(
                                labels[key], edit.replace(find, ""))
                  except: pass
        if self.next is not None: self.next.add_to(node)
        return node

    def add(self, what):
        print("add???", "%s" % self.data)
        if type(what) in (Suite, Family, Task):
            raise Exception(what.fullname())
        else:
            raise Exception(type(what))


class Variable(Edit): pass


class Variables(Edit): pass


class Repeat(Attribute):
    """ repeat date/int/string/enum """

    def __init__(self, name="YMD", start=20120101, end=21010101,
                 step=1, kind="date"):
        if kind in "date":
            self.data = ecflow.RepeatDate(name, int(start), int(end), int(step))
        elif "int" in kind:
            self.data = ecflow.RepeatInteger(
                name, int(start), int(end), int(step))
        elif kind == "string":
            self.data = ecflow.RepeatString(name, start)
        elif "enum" in kind:
            self.data = ecflow.RepeatEnumerated(name, start)
        elif kind == "day":
            self.data = ecflow.RepeatDay(step)
        else:
            self.data = None

    def add_to(self, node):
        if self.data is not None:
            node.load.add_repeat(self.data)
        return node


def If(test=True, then=None, otow=None, *args):
    """ enable Task("t1").add(If(test=(1==1),
                                 then=Edit(ONE=1),
                                 otow=Edit(TWO=2)))
        appreciate that both branches are evaluated, using this If class
        ie there is no 'dead code' as it is with python language 'if' structure

        using If to distinguish od/rd mode request that both users share
        the variables (parameter.py) and ecf.py

        otow: on the other way?
        """
    if len(args) > 1:
        try:
            print(then.name())
            print(otow.name())
        except: pass
        # print(args[0].name())
        raise Exception("test",
                        test, "\nthen\n",
                        then, "\nelse",
                        otow, "arg", args)
    if test:
        return then
    return otow


class Root(object):  # from where Suite and Node derive
    """ generic tree node """

    def __init__(self): self.load = None  # to be filled with ecFlow item

    def real(self): 
        # print(type(self.load))
        return self.load

    def __get_attr__(self, attr): return getattr(self.load, attr)

    def get_parent(self): return self.load.get_parent()

    def __str__(self):
        return self.fullname()

    def __repr__(self):
        return "%s" % self.load

    def __eq__(self, node):
        if isinstance(self.load, ecflow.Node):
            return "%s == " % self + str(node)
        return False

    def __ne__(self, node):
        if isinstance(self.load, ecflow.Node):
            return "%s != " % self + str(node)
        return False

    def __and__(self, node):
        if isinstance(self.load, ecflow.Node):
            return "%s and " % self + str(node)
        return False

    def __or__(self, node):
        if isinstance(self.load, ecflow.Node):
            return "%s or " % self + str(node)
        return False

    def fullname(self):
        """ simple syntax """
        if isinstance(self.load, ecflow.Node):
            return self.load.get_abs_node_path()
        return str(self)

    def repeat(self, name="YMD", start=20120101, end=20321212, step=1,
               kind="date"):
        """ add repeat attribute"""
        if kind in ("date", "integer"):
            self.add(Repeat(name, start, end, step, kind))
        elif kind in ("string", "enumerated"):
            self.add(Repeat(name, start, kind=kind))
        elif kind == "day":
            self.add(Repeat(step, kind="day"))
        else: raise Exception
        return self

    def defstatus(self, kind):
        """ add defstatus attribute"""
        status = kind
        if type(kind) == str:
            kinds = {"suspended": ecflow.DState.suspended,
                     "aborted": ecflow.DState.aborted,
                     "complete": ecflow.DState.complete,
                     "active": ecflow.DState.active,
                     "submitted": ecflow.DState.submitted,
                     "unknown": ecflow.DState.unknown,
                     "queued": ecflow.DState.queued, }
            status = kinds[kind]
        elif type(kind) == ecflow.DState: pass
        else: raise Exception("defstatus?")
        self.load.add_defstatus(status)
        return self

    def __add__(self, item): self.add(item)

    def append(self, item=None, *args):
        """ we get compatible with list then """
        return self.add(item, args)

    def add(self, item=None, *args):
        """ add a task, a family or an attribute """
        if DEBUG: print(self.fullname(), item, args)

        if item is not None:
            if type(item) == tuple:
                for val in item:
                    self.add(val)
            elif type(item) == list:
                for val in item:
                    self.add(val)
            elif type(item) == str:
                raise Exception(item)
            else:
                item.add_to(self)

        if len(args) > 0:
            if type(args) == tuple:
                for val in args:
                    self.add(val)
            elif type(args) == list:
                for val in args:
                    self.add(val)
            else: raise Exception
            return self

        if not isinstance(self.load, ecflow.Node):
            raise Exception("please dont", type(self))

        return self

    def limit(self, name, size):
        """ add limit attribute"""
        if name is None:
            raise Exception
        self.load.add_limit(name, size)
        return self

    def inlimit(self, full_path):
        """ add inlimit attribute"""
        if not USE_LIMIT:
            return self

        try:
            path, name = full_path.split(":")
        except ValueError:
            raise ValueError('invalid limit: {}'.format(full_path))

        self.load.add_inlimit(name, path)
        return self

    # follow pyflow
    shape = None
    def draw_tree(self): dot = Dot(fullnames=False); self._tree(dot); return dot

    def _tree(self, dot):
      try: dot.edge(self.load.get_parent(), self)
      except: pass
      for n in self.load.nodes:
        if n.name() != '_': _tree(dot, n)
        # if n.names[0] != '_': n._tree(dot)

    def draw_graph(self):
        dot = Dot(); self._graph(dot); dot.save("test.gv"); return dot

    def _graph(self, dot):
      for n in self.load.nodes:
        if n.name() != '_': _tree(dot, n)
      # for n in self._nodes.values():
      #  if n.names[0] != '_': n._tree(dot)

    # @property
    def to_html(self):
        # from .html import HTMLWrapper
        return "%s" % HTMLWrapper("%s" % self)  # .generate_node()))

    def _repr_html_(self): return str(self.to_html())


def _tree(dot, node):
      if type(node) in (ecflow.Node, ecflow.Family, ecflow.Task,
                        ecflow.ecflow.Family, ecflow.ecflow.Task):
          dot.edge(node.get_parent(), node)
          for n in node.nodes:
              if n.name() != '_': _tree(dot, n)
      elif type(node) in (Node, Family, Task):
          dot.edge(node.load.parent, node)
          _tree(dot, node.load)
      else: raise Exception(type(node))
      #try: dot.edge(node.load.parent, self)
      #except: pass


class Node(Root): # from where Task and Family derive
    """ Node class is shared by family and task """

    def __enter__(self): return self
    def __exit__(self, exc_type, exc_val, exc_tb): pass  # return self

    def real(self): return self.load

    def events(self):
        # for i in self.load.events: yield i                
        # print(i)

        # out = [x for x in self.load.events]; return out

        out = []
        for x in self.load.events: 
            out.append(x)
        print(out)
        return out

    def meters(self):
        # for i in self.load.meters: yield i        
        return [x for x in self.load.meters]

    #def nodes(self):
    # for i in self.load.nodes: yield i        
    #    return [x for x in self.load.nodes]

    def name(self):
        return self.load.name()

    def event(self, name=1):
        """ add event attribute"""
        if USE_EVENT:
            self.load.add_event(name)
        return self

    def meter(self, name, start, end, threshold=None):
        """ add meter attribute"""
        if threshold is None:
            threshold = end
        self.load.add_meter(name, start, end, threshold)
        return self

    def label(self, name, default=""):
        """ add label attribute"""
        self.load.add_label(name, default)
        return self

    def edit(self, name, value=""):
        """ add variable attribute"""
        self.load.add_variable(name, value)
        return self

    def variable(self, name, value=""): return self.edit(name, value)

    def cron(self, time, dom=False, wdays=False, month=False):
        """ wrapper for add_cron """
        cron = ecflow.Cron()
        cron.set_time_series(time)
        if wdays is not False:
            cron.set_week_days(wdays)
        if month is not False:
            cron.set_months(month)
        if dom is not False:
            cron.set_day_of_month(dom)

        self.load.add_cron(cron)
        return self

    def complete(self, arg):
        """ add complete attribute"""
        if USE_TRIGGER and arg is not None:
            self.load.add_complete(arg)
        return self

    def complete_and(self, arg):
        """ append to existing complete"""
        if USE_TRIGGER and arg is not None:
            self.load.add_part_complete(ecflow.PartExpression(arg, True))
        return self

    def complete_or(self, arg):
        """ append to existing complete"""
        if USE_TRIGGER and arg is not None:
            self.load.add_part_complete(ecflow.PartExpression(arg, False))
        return self

    def up(self):
        """ get parent, one level up"""
        return self.load.get_parent()


class Defs(object):
    """ wrapper for the definition """

    def __init__(self):
        self.load = ecflow.Defs()

    def __get_attr__(self, attr): 
        return getattr(self.load, attr)

    def auto_add_externs(self, true): self.load.auto_add_externs(true)
    def check(self): self.load.check()
    def add_extern(self, path): self.load.add_extern(path)  # TODO list
    def suites(self): 
        out = []
        for suite in self.load.suites:  # TODO list
            out.append(suite.name())
        return out

    def add_suite(self, node): 
        if type(node) == Suite:
            self.load.add_suite(node.load)  # TODO list
        elif type(node) == ecflow.Suite:
            self.load.add_suite(node)  # TODO list
        else: raise Exception("what?")

    def __str__(self): return "%s" % self.load
    __repr__ = __str__

    def add(self, item):
        """ add suite """
        if type(item) == Suite:
            self.load.add_suite(item.load)
        #elif type(item) == Extern:  # obsolete, Extern uses global DEFS var
        #    for path in item.folk: self.load.add_extern(path)
        elif type(item) == tuple:
            for one in item:
                self.add(one)
        elif type(item) == list:
            for one in item:
                self.add(one)
        elif item is None:
            pass
        else:
            raise Exception("ERR:load addwhat?", type(item))
        return self

    def suite(self, name):
        """ add suite providing its name """
        suite = Suite(name)
        self.add(suite)
        return suite


class Client(object):
    """ wrapper around client """

    def __init__(self, host="localhost", port="31415"):
        if host is None: host = "localhost"
        self.clnt = ecflow.Client(host, port)
        if "@" in host:
            host, port = host.split("@")
            self.clnt.set_host_port(host, int(port))
        else:
            super(Client, self).__init__()
            self.clnt.set_host_port(host, int(port))
        self.host = host
        self.port = port

    def load(self, defs):
        if type(defs) == Defs:
            self.clnt.load(defs.load)
        elif type(defs) == ecflow.Defs:
            self.clnt.load(defs)
        else: raise Exception("defs: really?")

    def replace(self, path, defs=None, parent=True, force=False):
        if defs is None: self.clnt.replace(path, DEFS, parent, force)
        elif type(defs) is Defs:
            self.clnt.replace(path, defs.load, parent, force)
        else: self.clnt.replace(path, defs, parent, force)

    def suites(self): return self.clnt.suites()
    def ping(self): self.clnt.ping()

    def __str__(self):
        return "ecflow client %s@%s v%s" % (
            self.host, self.port, self.version())


class Suite(Root):
    """ wrapper for a suite """

    def __enter__(self): return self
    def __exit__(self, exc_type, exc_val, exc_tb): pass  # return self

    def __init__(self, name):
        self.load = ecflow.Suite(name)

    def family(self, name):
        """ add a family """
        if "%" in name: raise Exception(name)
        obsolete()
        fam = Family(str(name))
        self.add(fam)
        return fam

    def task(self, name):
        """ add a task """
        if "%" in name: raise Exception(name)
        obsolete()
        tsk = Task(name)
        self.add(tsk)
        return tsk

    # def __enter__(self): return self

    # def __exit__(self, *args): pass


class Family(Node, Attribute):
    """ wrapper around family """

    def __init__(self, name):
        self.load = ecflow.Family(name)

    def family(self, name):
        """ add a family """
        if "%" in name: raise Exception(name)
        obsolete()
        fam = Family(str(name))
        self.add(fam)
        return fam

    def task(self, name):
        """ add a task """
        if "%" in name: raise Exception(name)
        obsolete()
        tsk = Task(name)
        self.add(tsk)
        return tsk

    def add_to(self, node):
        if DEBUG: print("add: %s %s" % (node.name(), self.name()))
        parent = self.load.get_parent()
        if parent: raise Exception("already attached...",
                                   parent.name(),
                                   node.name(), self.name())
        node.load.add_family(self.load)

    # def __enter__(self): return self

    # def __exit__(self, *args): pass


class Task(Node, Attribute):
    """ wrapper around task """

    def __init__(self, name):
        self.load = ecflow.Task(name)

    # def __setattr__(self, key, val):
    #     # assert key.isupper()
    #     if key.isupper():
    #         key, val = translate(key, val)
    #         print(type(self.load), "%s" % self.load)
    #         self.load.add(ecflow.Variable(key, val))

    def add_to(self, node):
        node.load.add_task(self.load)

    def add_family(self, node):
        raise Exception(self.name(), node.name(), self.fullname())

    def add_task(self, node):
        raise Exception(node.name(), self.name())


def display(defs, fname=None):
    """ print defs"""
    if fname is not None:
        fop = open(fname, "w")
        print(defs, file=fop)


class TestEcf(unittest.TestCase):
    """ a test case aka use-case """

    def test_xxx(self):
        """ a test """

        suite = Suite("Suite").add(Clock("real"),
                                     Autocancel(3))
        suite.defstatus("suspended")

        fam = Family("Family")
        tsk = Task("Task")

        ft2 = Task("Fam")
        ft2.add_to(fam)

        tsk.VAR = "VALUE"  # edit VAR "VALUE"
        tsk.add(Late("-s 00:05 -c 01:00"))

        fam.add(tsk,
                (Task("1"), Task("2")),
                [Task("11"), Task("12")],
                Task("111"), Task("211"),
                Task("t2").add(Trigger(tsk == COMPLETE),
                               Late("-s 00:05 -a 01:00 -c 20:00"),
                               Time("01:00")),
                Task("t21").add(Cron("00:00 23:59 01:00")),
                Task("t22").add(Today("14:00")),
                Task("t23").add(Date("1.*.*")),
                Task("t24").add(Day("monday")),
            )

        fam.add(Task("t3").add(
                If(test=(1 == 1),
                   then=Edit(ADD_ONE=1),
                   otow=Edit(ADD_TWO=1)),

                If(test=(1 == 0),
                   then=Edit(ADD_ONE=0),
                   otow=Edit(ADD_TWO=0)),
                Trigger(tsk != ABORTED),
                Complete(tsk == COMPLETE))) # longer

        fam.add(
            Task("2t"),
            Task("t4").add(
                Trigger(tsk.name() != COMPLETE)),
            Late("-s 00:05 -c 01:00"),
            Edit(VAR="VALUE"),
            Task("t5").add(Trigger(["t4", "t3", "t2"])),
            Task("t6").add(Trigger("2t" == COMPLETE)),
            Task("t7").add(Trigger("2t eq complete")), )

        tsk.add(Limit("a_limit", 10),
                Inlimit("a_task:a_limit"),
                Meter("step", -1, 100),
                Label("info", "none"),
                Event(1),
                Event("a"),
                Defstatus("complete"))

        Extern("/oper/main:YMD")
        tsk.add(Edit({"A": "a", "B": "b"}))  # dict
        tsk.add(Edit(D="d", E="e"))  # list name=value,
        tsk.add(Edit("C", "c"))  # name, value
        suite.add(fam,
                  Family("main").add(
                      Repeat("YMD", 20200202, 20320202, 7, "date"),
                      Task("doer").add(Time("04:55"))))

        fam.family("fam").add(DefcompleteIf(True))

        defs = Defs()
        defs.add(suite)
        another = defs.suite("suite")
        another.defstatus("suspended")
        another.task("tn")
        afam = another.family("family")
        afam.task("t2n")

        display(defs, fname="test_ecf.tmp")
        suite.draw_tree()
        suite.draw_graph()
        suite.to_html()

        print("%s" % suite)  # print name
        print(suite)  # print name
        print(repr(suite))  # print content
        print(defs)  # print content
        cmd = "xdg-open test.gv.pdf;"
        cmd += "dot -Tps test.gv > test.ps && xdg-open test.ps"; os.system(cmd)


# from pyflow, COMPAT
def definition_to_html(d):
    result = []
    for n in d.split("\n"): result.append(n)
    return "<pre>%s</pre>" % ("\n".join(result),)


class HTMLWrapper(object):
    def __init__(self, d): self._def = d
    def _repr_html_(self): return definition_to_html(self._def)
    __str__ = _repr_html_


def shapes(node):
    if type(node) in (ecflow.Task, Task):
        return "triangle"
    elif type(node) in (ecflow.Family, Family):
        return "rect"
    elif type(node) in (ecflow.Suite, Suite):
        return "invtriangle"
    return None


class Dot(object):
  """ follow tracks from pyflow """

  def __init__(self, fullnames=True):
      from graphviz import Digraph
      self._dot = Digraph()
      self._nodes = {}
      self._fullnames = fullnames

  def edge(self, node1, node2):
      self._dot.edge(self.node(node1), self.node(node2))

  def node(self, node):
      full = node.get_abs_node_path()
      if full not in self._nodes:
          self._nodes[full] = '%s' % node.name()
      self._dot.node(self._nodes[full], 
                     # fillcolor="shape",
                     # fontcolor="blue",
                     # fontsize=32,
                     # width=0.5,
                     # style="filled",
                     # fixedsize="shape",
                     # fixedsize="true",
                     # label="xxx",
                     shape=shapes(node))
      return self._nodes[full]

  def save(self, path, view=True):
      if os.path.exists(path): os.unlink(path)
      self._dot.render(path, view=view)

  # For jupyter notbeook
  def _repr_svg_(self):
      return self._dot._repr_svg_()


if __name__ == '__main__':
    unittest.main()
